# coding: utf8
from typing import Optional

"""This module contains utilities for DWI handling."""


def merge_volumes_tdim(in_file1, in_file2):
    """
    Merge 'in_file1' and 'in_file2' in the t dimension.

    Args:
        in_file1 (str): First set of volumes.
        in_file2 (str): Second set of volumes.

    Returns:
        out_file (str): The two sets of volumes merged.
    """
    import os
    import os.path as op

    out_file = op.abspath("merged_files.nii.gz")
    cmd = f"fslmerge -t {out_file} {in_file1} {in_file2}"
    os.system(cmd)
    return out_file


def count_b0s(in_bval, low_bval=5.0):
    """
    Count the number of volumes where b<=low_bval.

    Args:
        in_bval (str): bval file.
        low_bval (Optional[int]): Define the b0 volumes as all volume
            bval <= lowbval. (Default=5.0)

    Returns:
        num_b0s: Number of b0s.
    """
    import numpy as np

    bvals = np.loadtxt(in_bval)
    num_b0s = len(np.where(bvals <= low_bval)[0])

    return num_b0s


def b0_average(in_file: str, out_file: Optional[str] = None):
    """
    Average the b0 volumes.

    Args:
        in_file (str): The b0 volumes already registered.
        out_file (optional[str]): Name of the file.

    Returns:
        The mean of the b0 volumes.

    Warnings:
        The b0 volumes must be registered.
    """
    import os.path as op

    import nibabel as nb
    import numpy as np

    if not out_file:
        fname, ext = op.splitext(op.basename(in_file))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath("%s_avg_b0%s" % (fname, ext))

    imgs = np.array(nb.four_to_three(nb.load(in_file)))
    b0s = [im.get_fdata(dtype="float32").astype(np.float32) for im in imgs]
    b0 = np.average(np.array(b0s), axis=0)

    hdr = imgs[0].get_header().copy()
    hdr.set_data_shape(b0.shape)
    hdr.set_xyzt_units("mm")
    hdr.set_data_dtype(np.float32)
    nb.Nifti1Image(b0, imgs[0].get_affine(), hdr).to_filename(out_file)

    return out_file


def b0_dwi_split(in_dwi, in_bval, in_bvec, low_bval=5.0):
    """
    Split the DWI volumes into two datasets :
     - the first dataset contains the set of b<=low_bval volumes.
     - the second dataset contains the set of DWI volumes.

    Args:
        in_dwi (str): DWI dataset.
        in_bval (str): File describing the b-values of the DWI dataset.
        in_bvec (str): File describing the directions of the DWI dataset.
        low_bval (Optional[int]): Define the b0 volumes as all volume
            bval <= lowbval. (Default=5.0)

    Returns:
        out_b0 (str): The set of b<=low_bval volumes.
        out_dwi (str): Output. The set of b>low_bval volumes.
        out_bvals (str): The b-values corresponding to the out_dwi.
        out_bvecs (str): The b-vecs corresponding to the out_dwi.
    """
    import os.path as op
    import warnings

    import nibabel as nib
    import numpy as np

    assert op.isfile(in_dwi)
    assert op.isfile(in_bval)
    assert op.isfile(in_bvec)
    assert low_bval >= 0

    im = nib.load(in_dwi)
    data = im.get_fdata(dtype="float32")
    hdr = im.get_header().copy()
    bvals = np.loadtxt(in_bval)
    bvecs = np.loadtxt(in_bvec)

    if bvals.shape[0] == bvecs.shape[0]:
        warnings.warn(
            "Warning: The b-vectors file should be column-wise. "
            + "The b-vectors will be transposed",
            UserWarning,
        )
        bvecs = bvecs.T

    lowbs = np.where(bvals <= low_bval)[0]

    fname_b0, ext_b0 = op.splitext(op.basename(in_dwi))
    if ext_b0 == ".gz":
        fname_b0, ext2 = op.splitext(fname_b0)
        ext_b0 = ext2 + ext_b0
    out_b0 = op.abspath("%s_b0%s" % (fname_b0, ext_b0))
    # out_b0 = op.abspath('b0.nii.gz')
    b0data = data[..., lowbs]
    hdr.set_data_shape(b0data.shape)
    nib.Nifti1Image(b0data, im.get_affine(), hdr).to_filename(out_b0)

    dwi_bvals = np.where(bvals > low_bval)[0]
    out_dwi = op.abspath("dwi.nii.gz")
    dwi_data = data[..., dwi_bvals]
    hdr.set_data_shape(dwi_data.shape)
    nib.Nifti1Image(dwi_data, im.get_affine(), hdr).to_filename(out_dwi)

    bvals_dwi = bvals[dwi_bvals]
    out_bvals = op.abspath("bvals")
    np.savetxt(out_bvals, bvals_dwi, fmt="%d", delimiter=" ")

    bvecs_dwi = np.array(
        [
            bvecs[0][dwi_bvals].tolist(),
            bvecs[1][dwi_bvals].tolist(),
            bvecs[2][dwi_bvals].tolist(),
        ]
    )
    out_bvecs = op.abspath("bvecs")
    np.savetxt(out_bvecs, bvecs_dwi, fmt="%10.5f", delimiter=" ")

    return out_b0, out_dwi, out_bvals, out_bvecs


def insert_b0_into_dwi(in_b0, in_dwi, in_bval, in_bvec):
    """
    This function inserts a b0 volume into the dwi dataset as the first volume
    and updates the bvals and bvecs files.

    Args:
        in_b0 (str): One b=0 volume (could be the average of a b0 dataset).
        in_dwi (str): The set of DWI volumes.
        in_bval (str): File describing the b-values of the DWI dataset.
        in_bvec (str): File describing the directions of the DWI dataset.

    Returns:
        out_dwi (str): Diffusion dataset : b0 volume + dwi volumes.
        out_bval (str): B-values update.
        out_bvec (str): Directions of diffusion update.
    """
    import os.path as op

    import numpy as np

    from clinica.utils.dwi import merge_volumes_tdim

    assert op.isfile(in_b0)
    assert op.isfile(in_dwi)
    assert op.isfile(in_bval)
    assert op.isfile(in_bvec)

    out_dwi = merge_volumes_tdim(in_b0, in_dwi)

    lst = np.loadtxt(in_bval).tolist()
    lst.insert(0, 0)
    out_bvals = op.abspath("bvals")
    np.savetxt(out_bvals, np.matrix(lst), fmt="%d", delimiter=" ")

    bvecs = np.loadtxt(in_bvec)
    bvecs_0 = bvecs[0].tolist()
    bvecs_0.insert(0, 0.0)
    bvecs_1 = bvecs[1].tolist()
    bvecs_1.insert(0, 0.0)
    bvecs_2 = bvecs[2].tolist()
    bvecs_2.insert(0, 0.0)
    bvecs_dwi = np.array([bvecs_0, bvecs_1, bvecs_2])
    out_bvecs = op.abspath("bvecs")
    np.savetxt(out_bvecs, bvecs_dwi, fmt="%10.5f", delimiter=" ")

    return out_dwi, out_bvals, out_bvecs


def prepare_reference_b0(in_dwi, in_bval, in_bvec, low_bval=5):
    """
    This function prepare the data for further corrections. It coregisters
    the B0 images and then average it in order to obtain only
    one average B0 images.

    Args:
        in_dwi (str): Input DWI file.
        in_bvec (str): Vector file of the diffusion directions of
            the dwi dataset.
        in_bval (str): B-values file.
        low_bval (optional[int]):

    Returns:
        out_reference_b0 (str): Average of the B0 images or the only B0 image.
        out_b0_mask (str): Binary mask obtained from the average of
            the B0 images.
        out_b0_dwi_merge (str): Average of B0 images merged to the DWIs.
        out_updated_bval (str): Updated gradient values table.
        out_updated_bvec (str): Updated gradient vectors table.
    """
    import os.path as op
    import tempfile

    from clinica.utils.dwi import (
        b0_average,
        b0_dwi_split,
        count_b0s,
        insert_b0_into_dwi,
    )
    from clinica.utils.stream import cprint
    from clinica.workflows.dwi_preprocessing import b0_flirt_pipeline

    # Count the number of b0s
    nb_b0s = count_b0s(in_bval=in_bval, low_bval=low_bval)
    # cprint('Found %s b0 for %s' % (nb_b0s, in_dwi))

    # Split dataset into two datasets: the b0 and the b>low_bval datasets
    [extracted_b0, out_split_dwi, out_split_bval, out_split_bvec] = b0_dwi_split(
        in_dwi=in_dwi, in_bval=in_bval, in_bvec=in_bvec, low_bval=low_bval
    )

    if nb_b0s == 1:
        # The reference b0 is the extracted b0
        # cprint('Only one b0 for %s' % in_dwi)
        out_reference_b0 = extracted_b0
    elif nb_b0s > 1:
        # Register the b0 onto the first b0
        b0_flirt = b0_flirt_pipeline(num_b0s=nb_b0s)
        b0_flirt.inputs.inputnode.in_file = extracted_b0
        # BUG: Nipype does allow to extract the output after running the
        # workflow: we need to 'guess' where the output will be generated
        tmp_dir = tempfile.mkdtemp()
        b0_flirt.base_dir = tmp_dir
        b0_flirt.run()
        # out_node = b0_flirt.get_node('outputnode')
        registered_b0s = op.abspath(
            op.join(
                tmp_dir, "b0_coregistration", "concat_ref_moving", "merged_files.nii.gz"
            )
        )
        # cprint('B0 s will be averaged (file = ' + registered_b0s + ')')
        # Average the b0s to obtain the reference b0
        out_reference_b0 = b0_average(in_file=registered_b0s)
    else:
        raise ValueError(
            "The number of b0s should be strictly positive (b-val file =%s)." % in_bval
        )

    # Merge datasets such that bval(DWI) = (0 b1 ... bn)
    [out_b0_dwi_merge, out_updated_bval, out_updated_bvec] = insert_b0_into_dwi(
        in_b0=out_reference_b0,
        in_dwi=out_split_dwi,
        in_bval=out_split_bval,
        in_bvec=out_split_bvec,
    )

    return out_reference_b0, out_b0_dwi_merge, out_updated_bval, out_updated_bvec


def hmc_split(in_file, in_bval, ref_num=0, lowbval=5.0):
    """
    Selects the reference ('out_ref') and moving ('out_mov') volumes
    from a dwi dataset for the purpose of head motion correction (HMC).

    Args:
        in_file (str): DWI dataset.
        in_bval (str): File describing the b-values of the DWI dataset.
        ref_num (Optional[str]): The reference volume in the dwi dataset.
            Default ref_num= 0.
        lowbval (Optional[int]): Define the b0 volumes as all volume
            bval <= lowbval. (Default=5.0)
    Returns:
        out_ref (str): The reference volume.
        out_mov (str): The moving volume to align to the reference volume.
        out_bval (str): The b-values corresponding to the moving volume.
        volid (int): Index of the reference volume.
    """
    import os.path as op

    import nibabel as nib
    import numpy as np

    im = nib.load(in_file)
    data = im.get_fdata(dtype="float32")
    hdr = im.get_header().copy()
    bval = np.loadtxt(in_bval)

    lowbs = np.where(bval <= lowbval)[0]
    assert ref_num in lowbs
    volid = ref_num

    out_ref = op.abspath("hmc_ref.nii.gz")
    refdata = data[..., volid]
    hdr.set_data_shape(refdata.shape)
    nib.Nifti1Image(refdata, im.get_affine(), hdr).to_filename(out_ref)

    if volid == 0:
        data = data[..., 1:]
        bval = bval[1:]
    elif volid == (data.shape[-1] - 1):
        data = data[..., :-1]
        bval = bval[:-1]
    else:
        data = np.concatenate((data[..., :volid], data[..., (volid + 1) :]), axis=3)
        bval = np.hstack((bval[:volid], bval[(volid + 1) :]))

    out_mov = op.abspath("hmc_mov.nii.gz")
    out_bval = op.abspath("bval_split.txt")

    hdr.set_data_shape(data.shape)
    nib.Nifti1Image(data, im.get_affine(), hdr).to_filename(out_mov)
    np.savetxt(out_bval, bval)
    return out_ref, out_mov, out_bval, volid


def check_dwi_volume(in_dwi, in_bvec, in_bval):
    # TODO return str instead of raising an error, so that user can be informed of which file is causing problem
    """
    Check that # DWI = # B-val = # B-vec.

    Raises
        IOError
    """
    import nibabel as nib
    import numpy as np

    bvals = np.loadtxt(in_bval)
    num_b_vals = len(bvals)

    bvecs = np.loadtxt(in_bvec)
    _, num_b_vecs = bvecs.shape

    img = nib.load(in_dwi)
    _, _, _, num_dwis = img.shape

    if not (num_b_vals == num_b_vecs == num_dwis):
        raise IOError(
            f"Number of DWIs, b-vals and b-vecs mismatch "
            f"(# DWI = {num_dwis}, # B-vec = {num_b_vecs}, #B-val = {num_b_vals}) "
        )



from typing import List, Optional

import click

from clinica.pipelines import cli_param

pipeline_name = "t1-volume-register-dartel"


@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.tissues
def cli(
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    tissues: List[int] = (1, 2, 3),
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Inter-subject registration using Dartel (using an existing Dartel template).

       GROUP_LABEL is an user-defined identifier to target a specific group of subjects. For this pipeline, it is associated to the DARTEL template that you had created when running the t1-volume pipeline.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_Volume/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .t1_volume_register_dartel_pipeline import T1VolumeRegisterDartel

    parameters = {"group_label": group_label, "tissues": tissues}

    pipeline = T1VolumeRegisterDartel(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



# coding: utf8

import clinica.pipelines.engine as cpe


class T1VolumeDartel2MNI(cpe.Pipeline):
    """T1VolumeDartel2MNI - Dartel template to MNI.

    Returns:
        A clinica pipeline object containing the T1VolumeDartel2MNI pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        if "group_label" not in self.parameters.keys():
            raise KeyError("Missing compulsory group_label key in pipeline parameter.")

        self.parameters.setdefault("tissues", [1, 2, 3])
        self.parameters.setdefault("voxel_size", None)
        self.parameters.setdefault("modulate", True)
        self.parameters.setdefault("smooth", [8])

        check_group_label(self.parameters["group_label"])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ["native_segmentations", "flowfield_files", "template_file"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return ["normalized_files", "smoothed_normalized_files", "atlas_statistics"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.input_files import (
            t1_volume_deformation_to_template,
            t1_volume_final_group_template,
            t1_volume_native_tpm,
        )
        from clinica.utils.inputs import clinica_file_reader, clinica_group_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import (
            print_groups_in_caps_directory,
            print_images_to_process,
        )

        # Check that group already exists
        if not os.path.exists(
            os.path.join(
                self.caps_directory, "groups", "group-" + self.parameters["group_label"]
            )
        ):
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                f"Group {self.parameters['group_label']} does not exist. "
                "Did you run t1-volume or t1-volume-create-dartel pipeline?"
            )

        all_errors = []
        read_input_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )

        # Segmented Tissues
        # =================
        tissues_input = []
        for tissue_number in self.parameters["tissues"]:
            try:
                native_space_tpm = clinica_file_reader(
                    self.subjects,
                    self.sessions,
                    self.caps_directory,
                    t1_volume_native_tpm(tissue_number),
                )
                tissues_input.append(native_space_tpm)
            except ClinicaException as e:
                all_errors.append(e)
        # Tissues_input has a length of len(self.parameters['mask_tissues']). Each of these elements has a size of
        # len(self.subjects). We want the opposite : a list of size len(self.subjects) whose elements have a size of
        # len(self.parameters['mask_tissues']. The trick is to iter on elements with zip(*my_list)
        tissues_input_rearranged = []
        for subject_tissue_list in zip(*tissues_input):
            tissues_input_rearranged.append(subject_tissue_list)

        read_input_node.inputs.native_segmentations = tissues_input_rearranged

        # Flow Fields
        # ===========
        try:
            read_input_node.inputs.flowfield_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                t1_volume_deformation_to_template(self.parameters["group_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        # Dartel Template
        # ================
        try:
            read_input_node.inputs.template_file = clinica_group_reader(
                self.caps_directory,
                t1_volume_final_group_template(self.parameters["group_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = "Clinica faced error(s) while trying to read files in your CAPS/BIDS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaCAPSError(error_message)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint("The pipeline will last a few minutes per image.")

        # fmt: off
        self.connect(
            [
                (read_input_node, self.input_node, [("native_segmentations", "native_segmentations"),
                                                    ("flowfield_files", "flowfield_files"),
                                                    ("template_file", "template_file")]),
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import zip_nii

        # Writing normalized images (and smoothed) into CAPS
        # ==================================================
        write_normalized_node = npe.MapNode(
            name="write_normalized_node",
            iterfield=["container", "normalized_files", "smoothed_normalized_files"],
            interface=nio.DataSink(
                infields=["normalized_files", "smoothed_normalized_files"]
            ),
        )
        write_normalized_node.inputs.base_directory = self.caps_directory
        write_normalized_node.inputs.parameterization = False
        write_normalized_node.inputs.container = [
            "subjects/"
            + self.subjects[i]
            + "/"
            + self.sessions[i]
            + "/t1/spm/dartel/group-"
            + self.parameters["group_label"]
            for i in range(len(self.subjects))
        ]
        write_normalized_node.inputs.regexp_substitutions = [
            (r"(.*)c1(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-graymatter_probability\3"),
            (r"(.*)c2(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-whitematter_probability\3"),
            (r"(.*)c3(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-csf_probability\3"),
            (r"(.*)c4(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-bone_probability\3"),
            (r"(.*)c5(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-softtissue_probability\3"),
            (r"(.*)c6(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-background_probability\3"),
            (
                r"(.*)mw(sub-.*)_probability(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_modulated-on_probability\3",
            ),
            (
                r"(.*)w(sub-.*)_probability(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_modulated-off_probability\3",
            ),
            (r"(.*)/normalized_files/(sub-.*)$", r"\1/\2"),
            (
                r"(.*)/smoothed_normalized_files/(fwhm-[0-9]+mm)_(sub-.*)_probability(\.nii(\.gz)?)$",
                r"\1/\3_\2_probability\4",
            ),
            (r"trait_added", r""),
        ]

        # fmt: off
        self.connect(
            [
                (self.output_node, write_normalized_node,
                    [
                        (("normalized_files", zip_nii, True), "normalized_files"),
                        (("smoothed_normalized_files", zip_nii, True), "smoothed_normalized_files"),
                    ],
                )
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.spm as spm
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import unzip_nii
        from clinica.utils.spm import spm_standalone_is_available, use_spm_standalone

        from ..t1_volume_dartel2mni import (
            t1_volume_dartel2mni_utils as dartel2mni_utils,
        )

        if spm_standalone_is_available():
            use_spm_standalone()

        # Unzipping
        # =========
        unzip_tissues_node = npe.MapNode(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_tissues_node",
            iterfield=["in_file"],
        )
        unzip_flowfields_node = npe.MapNode(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_flowfields_node",
            iterfield=["in_file"],
        )
        unzip_template_node = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_template_node",
        )

        # DARTEL2MNI Registration
        # =======================
        dartel2mni_node = npe.MapNode(
            spm.DARTELNorm2MNI(),
            name="dartel2MNI",
            iterfield=["apply_to_files", "flowfield_files"],
        )
        if self.parameters["voxel_size"] is not None:
            dartel2mni_node.inputs.voxel_size = tuple(self.parameters["voxel_size"])
        dartel2mni_node.inputs.modulate = self.parameters["modulate"]
        dartel2mni_node.inputs.fwhm = 0

        # Smoothing
        # =========
        if self.parameters["smooth"] is not None and len(self.parameters["smooth"]) > 0:
            smoothing_node = npe.MapNode(
                spm.Smooth(), name="smoothing_node", iterfield=["in_files"]
            )

            smoothing_node.iterables = [
                ("fwhm", [[x, x, x] for x in self.parameters["smooth"]]),
                (
                    "out_prefix",
                    ["fwhm-" + str(x) + "mm_" for x in self.parameters["smooth"]],
                ),
            ]
            smoothing_node.synchronize = True

            join_smoothing_node = npe.JoinNode(
                interface=nutil.Function(
                    input_names=["smoothed_normalized_files"],
                    output_names=["smoothed_normalized_files"],
                    function=dartel2mni_utils.join_smoothed_files,
                ),
                joinsource="smoothing_node",
                joinfield="smoothed_normalized_files",
                name="join_smoothing_node",
            )
            # fmt: off
            self.connect(
                [
                    (dartel2mni_node, smoothing_node, [("normalized_files", "in_files")]),
                    (smoothing_node, join_smoothing_node, [("smoothed_files", "smoothed_normalized_files")]),
                    (join_smoothing_node, self.output_node, [("smoothed_normalized_files", "smoothed_normalized_files")]),
                ]
            )
            # fmt: on
        else:
            self.output_node.inputs.smoothed_normalized_files = []

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, unzip_tissues_node, [("native_segmentations", "in_file")]),
                (self.input_node, unzip_flowfields_node, [("flowfield_files", "in_file")]),
                (self.input_node, unzip_template_node, [("template_file", "in_file")]),
                (unzip_tissues_node, dartel2mni_node, [("out_file", "apply_to_files")]),
                (unzip_flowfields_node, dartel2mni_node, [
                    (("out_file", dartel2mni_utils.prepare_flowfields, self.parameters["tissues"]), "flowfield_files")]),
                (unzip_template_node, dartel2mni_node, [("out_file", "template_file")]),
                (dartel2mni_node, self.output_node, [("normalized_files", "normalized_files")]),
            ]
        )
        # fmt: on



# coding: utf8

"""T1 Linear - Clinica Visualizer.
This file has been generated automatically by the `clinica generate template`
command line tool. See here for more details:
https://aramislab.paris.inria.fr/clinica/docs/public/latest/InteractingWithClinica/
"""



# coding: utf8


import itertools
import json
from multiprocessing.pool import ThreadPool
from os import path

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import StratifiedKFold
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from sklearn.svm import SVC
from xgboost import XGBClassifier

import clinica.pipelines.machine_learning.ml_utils as utils
from clinica.pipelines.machine_learning import base


class DualSVMAlgorithm(base.MLAlgorithm):
    def _launch_svc(self, kernel_train, x_test, y_train, y_test, c):

        if self._algorithm_params["balanced"]:
            svc = SVC(
                C=c,
                kernel="precomputed",
                probability=True,
                tol=1e-6,
                class_weight="balanced",
            )
        else:
            svc = SVC(C=c, kernel="precomputed", probability=True, tol=1e-6)

        svc.fit(kernel_train, y_train)
        y_hat_train = svc.predict(kernel_train)
        y_hat = svc.predict(x_test)
        proba_test = svc.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return svc, y_hat, auc, y_hat_train

    def _grid_search(self, kernel_train, x_test, y_train, y_test, c):

        _, y_hat, _, _ = self._launch_svc(kernel_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction(y_test, y_hat)

        return res["balanced_accuracy"]

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {"c": best_c, "balanced_accuracy": best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params["n_threads"])
        async_result = {}
        for i in range(self._algorithm_params["grid_search_folds"]):
            async_result[i] = {}

        outer_kernel = self._kernel[train_index, :][:, train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(
            n_splits=self._algorithm_params["grid_search_folds"], shuffle=True
        )
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            inner_kernel = outer_kernel[inner_train_index, :][:, inner_train_index]
            x_test_inner = outer_kernel[inner_test_index, :][:, inner_train_index]
            y_train_inner, y_test_inner = (
                y_train[inner_train_index],
                y_train[inner_test_index],
            )

            for c in self._algorithm_params["c_range"]:
                async_result[i][c] = inner_pool.apply_async(
                    self._grid_search,
                    (inner_kernel, x_test_inner, y_train_inner, y_test_inner, c),
                )
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._kernel[test_index, :][:, train_index]
        y_train, y_test = self._y[train_index], self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_svc(
            outer_kernel, x_test, y_train, y_test, best_parameter["c"]
        )

        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction(y_test, y_hat)
        result["evaluation_train"] = utils.evaluate_prediction(y_train, y_hat_train)
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index
        result["auc"] = auc

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result["best_parameter"]["c"])
            bal_acc_list.append(result["best_parameter"]["balanced_accuracy"])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params["balanced"]:
            svc = SVC(
                C=best_c,
                kernel="precomputed",
                probability=True,
                tol=1e-6,
                class_weight="balanced",
            )
        else:
            svc = SVC(C=best_c, kernel="precomputed", probability=True, tol=1e-6)

        svc.fit(self._kernel, self._y)

        return svc, {"c": best_c, "balanced_accuracy": mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(
            path.join(output_dir, "dual_coefficients.txt"), classifier.dual_coef_
        )
        np.savetxt(
            path.join(output_dir, "support_vectors_indices.txt"), classifier.support_
        )
        np.savetxt(path.join(output_dir, "intersect.txt"), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        dual_coefficients = classifier.dual_coef_
        sv_indices = classifier.support_

        weighted_sv = dual_coefficients.transpose() * x[sv_indices]
        weights = np.sum(weighted_sv, 0)

        np.savetxt(path.join(output_dir, "weights.txt"), weights)

        return weights

    def save_parameters(self, parameters_dict, output_dir):
        with open(path.join(output_dir, "best_parameters.json"), "w") as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return True

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "balanced": True,
            "grid_search_folds": 10,
            "c_range": np.logspace(-6, 2, 17),
            "n_threads": 15,
        }

        return parameters_dict


class LogisticReg(base.MLAlgorithm):
    def _launch_logistic_reg(self, x_train, x_test, y_train, y_test, c):

        if self._algorithm_params["balanced"]:
            classifier = LogisticRegression(
                penalty=self._algorithm_params["penalty"],
                tol=1e-6,
                C=c,
                class_weight="balanced",
            )
        else:
            classifier = LogisticRegression(
                penalty=self._algorithm_params["penalty"], tol=1e-6, C=c
            )

        classifier.fit(x_train, y_train)
        y_hat_train = classifier.predict(x_train)
        y_hat = classifier.predict(x_test)
        proba_test = classifier.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return classifier, y_hat, auc, y_hat_train

    def _grid_search(self, x_train, x_test, y_train, y_test, c):

        _, y_hat, _, _ = self._launch_logistic_reg(x_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction(y_test, y_hat)

        return res["balanced_accuracy"]

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {"c": best_c, "balanced_accuracy": best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params["n_threads"])
        async_result = {}
        for i in range(self._algorithm_params["grid_search_folds"]):
            async_result[i] = {}

        x_train = self._x[train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(
            n_splits=self._algorithm_params["grid_search_folds"], shuffle=True
        )
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            x_train_inner = x_train[inner_train_index]
            x_test_inner = x_train[inner_test_index]
            y_train_inner = y_train[inner_train_index]
            y_test_inner = y_train[inner_test_index]

            for c in self._algorithm_params["c_range"]:
                async_result[i][c] = inner_pool.apply_async(
                    self._grid_search,
                    (x_train_inner, x_test_inner, y_train_inner, y_test_inner, c),
                )
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_logistic_reg(
            x_train, x_test, y_train, y_test, best_parameter["c"]
        )

        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction(y_test, y_hat)
        result["evaluation_train"] = utils.evaluate_prediction(y_train, y_hat_train)
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index
        result["auc"] = auc

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result["best_parameter"]["c"])
            bal_acc_list.append(result["best_parameter"]["balanced_accuracy"])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params["balanced"]:
            classifier = LogisticRegression(
                C=best_c,
                penalty=self._algorithm_params["penalty"],
                tol=1e-6,
                class_weight="balanced",
            )
        else:
            classifier = LogisticRegression(
                C=best_c, penalty=self._algorithm_params["penalty"], tol=1e-6
            )

        classifier.fit(self._x, self._y)

        return classifier, {"c": best_c, "balanced_accuracy": mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(path.join(output_dir, "weights.txt"), classifier.coef_.transpose())
        np.savetxt(path.join(output_dir, "intercept.txt"), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        np.savetxt(path.join(output_dir, "weights.txt"), classifier.coef_.transpose())
        return classifier.coef_.transpose()

    def save_parameters(self, parameters_dict, output_dir):

        with open(path.join(output_dir, "best_parameters.json"), "w") as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def _centered_normalised_data(features):
        std = np.std(features, axis=0)
        std[np.where(std == 0)[0]] = 1.0
        mean = np.mean(features, axis=0)
        features_bis = (features - mean) / std
        return features_bis, mean, std

    @staticmethod
    def uses_kernel():
        return False

    @staticmethod
    def get_default_parameters():
        parameters_dict = {
            "penalty": "l2",
            "balanced": False,
            "grid_search_folds": 10,
            "c_range": np.logspace(-6, 2, 17),
            "n_threads": 15,
        }

        return parameters_dict


class RandomForest(base.MLAlgorithm):
    def _launch_random_forest(
        self,
        x_train,
        x_test,
        y_train,
        y_test,
        n_estimators,
        max_depth,
        min_samples_split,
        max_features,
    ):

        if self._algorithm_params["balanced"]:
            classifier = RandomForestClassifier(
                n_estimators=n_estimators,
                max_depth=max_depth,
                min_samples_split=min_samples_split,
                max_features=max_features,
                class_weight="balanced",
                n_jobs=self._algorithm_params["n_threads"],
            )
        else:
            classifier = RandomForestClassifier(
                n_estimators=n_estimators,
                max_depth=max_depth,
                min_samples_split=min_samples_split,
                max_features=max_features,
                n_jobs=self._algorithm_params["n_threads"],
            )

        classifier.fit(x_train, y_train)
        y_hat_train = classifier.predict(x_train)
        y_hat = classifier.predict(x_test)
        proba_test = classifier.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return classifier, y_hat, auc, y_hat_train

    def _grid_search(
        self,
        x_train,
        x_test,
        y_train,
        y_test,
        n_estimators,
        max_depth,
        min_samples_split,
        max_features,
    ):

        _, y_hat, _, _ = self._launch_random_forest(
            x_train,
            x_test,
            y_train,
            y_test,
            n_estimators,
            max_depth,
            min_samples_split,
            max_features,
        )
        res = utils.evaluate_prediction(y_test, y_hat)

        return res["balanced_accuracy"]

    def _select_best_parameter(self, async_result):

        params_list = []
        accuracies = []

        all_params_acc = []

        for fold in async_result.keys():
            best_params = None
            best_acc = -1

            for params, async_acc in async_result[fold].items():
                acc = async_acc.get()
                if acc > best_acc:
                    best_params = params
                    best_acc = acc

                all_params_acc.append(
                    pd.DataFrame(
                        {
                            "n_estimators": params[0],
                            "max_depth": params[1],
                            "min_samples_split": params[2],
                            "max_features": params[3],
                            "balanced_accuracy": acc,
                        },
                        index=[
                            "i",
                        ],
                    )
                )

            params_list.append(best_params)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_n_estimators = int(round(np.mean([x[0] for x in params_list])))
        best_max_depth = int(
            round(np.mean([x[1] if x[1] is not None else 50 for x in params_list]))
        )
        best_min_samples_split = int(round(np.mean([x[2] for x in params_list])))

        def max_feature_to_float(m):
            if type(m) is float:
                return m
            if type(m) is int:
                return float(m) / float(self._x.shape[1])
            if m == "auto" or m == "sqrt":
                return np.sqrt(self._x.shape[1]) / float(self._x.shape[1])
            if m == "log2":
                return np.log2(self._x.shape[1]) / float(self._x.shape[1])
            raise ValueError("Not valid value for max_feature: %s" % m)

        float_max_feat = [max_feature_to_float(x[3]) for x in params_list]
        best_max_features = np.mean(float_max_feat)

        return {
            "n_estimators": best_n_estimators,
            "max_depth": best_max_depth,
            "min_samples_split": best_min_samples_split,
            "max_features": best_max_features,
            "balanced_accuracy": best_acc,
        }

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params["n_threads"])
        async_result = {}
        for i in range(self._algorithm_params["grid_search_folds"]):
            async_result[i] = {}

        x_train = self._x[train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(
            n_splits=self._algorithm_params["grid_search_folds"], shuffle=True
        )
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        parameters_combinations = list(
            itertools.product(
                self._algorithm_params["n_estimators_range"],
                self._algorithm_params["max_depth_range"],
                self._algorithm_params["min_samples_split_range"],
                self._algorithm_params["max_features_range"],
            )
        )

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            x_train_inner = x_train[inner_train_index]
            x_test_inner = x_train[inner_test_index]
            y_train_inner = y_train[inner_train_index]
            y_test_inner = y_train[inner_test_index]

            for parameters in parameters_combinations:
                async_result[i][parameters] = inner_pool.apply_async(
                    self._grid_search,
                    (
                        x_train_inner,
                        x_test_inner,
                        y_train_inner,
                        y_test_inner,
                        parameters[0],
                        parameters[1],
                        parameters[2],
                        parameters[3],
                    ),
                )
        inner_pool.close()
        inner_pool.join()
        best_parameter = self._select_best_parameter(async_result)
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_random_forest(
            x_train,
            x_test,
            y_train,
            y_test,
            best_parameter["n_estimators"],
            best_parameter["max_depth"],
            best_parameter["min_samples_split"],
            best_parameter["max_features"],
        )

        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction(y_test, y_hat)
        result["evaluation_train"] = utils.evaluate_prediction(y_train, y_hat_train)
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index
        result["auc"] = auc

        return result

    def evaluate_no_cv(self, train_index, test_index):

        x_train = self._x[train_index]
        y_train = self._y[train_index]
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        best_parameter = dict()
        best_parameter["n_estimators"] = self._algorithm_params["n_estimators_range"]
        best_parameter["max_depth"] = self._algorithm_params["max_depth_range"]
        best_parameter["min_samples_split"] = self._algorithm_params[
            "min_samples_split_range"
        ]
        best_parameter["max_features"] = self._algorithm_params["max_features_range"]

        _, y_hat, auc, y_hat_train = self._launch_random_forest(
            x_train,
            x_test,
            y_train,
            y_test,
            self._algorithm_params["n_estimators_range"],
            self._algorithm_params["max_depth_range"],
            self._algorithm_params["min_samples_split_range"],
            self._algorithm_params["max_features_range"],
        )
        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction(y_test, y_hat)
        best_parameter["balanced_accuracy"] = result["evaluation"]["balanced_accuracy"]
        result["evaluation_train"] = utils.evaluate_prediction(y_train, y_hat_train)
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index
        result["auc"] = auc

        return result

    def apply_best_parameters(self, results_list):

        mean_bal_acc = np.mean(
            [result["best_parameter"]["balanced_accuracy"] for result in results_list]
        )
        best_n_estimators = int(
            round(
                np.mean(
                    [
                        result["best_parameter"]["n_estimators"]
                        for result in results_list
                    ]
                )
            )
        )
        best_max_depth = int(
            round(
                np.mean(
                    [
                        result["best_parameter"]["max_depth"]
                        if result["best_parameter"]["max_depth"] is not None
                        else 50
                        for result in results_list
                    ]
                )
            )
        )
        best_min_samples_split = int(
            round(
                np.mean(
                    [
                        result["best_parameter"]["min_samples_split"]
                        for result in results_list
                    ]
                )
            )
        )

        max_feat = []
        n_features = self._x.shape[1]
        for result in results_list:
            result_feat = result["best_parameter"]["max_features"]

            if not result_feat:
                max_features = 1.0
            elif result_feat in ["auto", "sqrt"]:
                max_features = np.sqrt(n_features) / n_features
            elif result_feat == "log2":
                max_features = np.log2(n_features) / n_features
            elif isinstance(result_feat, int):
                max_features = float(result_feat) / n_features
            elif isinstance(result_feat, float):
                max_features = result_feat
            else:
                raise ("Unknown max_features type")

            max_feat.append(max_features)
        best_max_features = np.mean(max_feat)

        if self._algorithm_params["balanced"]:
            classifier = RandomForestClassifier(
                n_estimators=best_n_estimators,
                max_depth=best_max_depth,
                min_samples_split=best_min_samples_split,
                max_features=best_max_features,
                class_weight="balanced",
                n_jobs=self._algorithm_params["n_threads"],
            )
        else:
            classifier = RandomForestClassifier(
                n_estimators=best_n_estimators,
                max_depth=best_max_depth,
                min_samples_split=best_min_samples_split,
                max_features=best_max_features,
                n_jobs=self._algorithm_params["n_threads"],
            )

        classifier.fit(self._x, self._y)

        return classifier, {
            "n_estimators": best_n_estimators,
            "max_depth": best_max_depth,
            "min_samples_split": best_min_samples_split,
            "max_features": best_max_features,
            "balanced_accuracy": mean_bal_acc,
        }

    def save_classifier(self, classifier, output_dir):
        np.savetxt(
            path.join(output_dir, "feature_importances.txt"),
            classifier.feature_importances_,
        )
        # print classifier.estimators_
        # np.savetxt(path.join(output_dir, 'estimators.txt'), str(classifier.estimators_))

    def save_weights(self, classifier, x, output_dir):

        np.savetxt(
            path.join(output_dir, "weights.txt"), classifier.feature_importances_
        )
        return classifier.feature_importances_

    def save_parameters(self, parameters_dict, output_dir):

        with open(path.join(output_dir, "best_parameters.json"), "w") as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return False

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "balanced": False,
            "grid_search_folds": 10,
            "n_estimators_range": (10, 25, 50, 100, 150, 200, 500),
            "max_depth_range": (None, 6, 8, 10, 12),
            "min_samples_split_range": (2, 4, 6, 8),
            "max_features_range": ("auto", 0.1, 0.2, 0.3, 0.4, 0.5),
            "n_threads": 15,
        }

        return parameters_dict


class XGBoost(base.MLAlgorithm):
    def _launch_xgboost(
        self,
        x_train,
        x_test,
        y_train,
        y_test,
        max_depth,
        learning_rate,
        n_estimators,
        colsample_bytree,
    ):

        if self._algorithm_params["balanced"]:
            # set scale_pos_weight
            # http://xgboost.readthedocs.io/en/latest//how_to/param_tuning.html
            scale_pos_weight = float(len(self._y - sum(self._y)) / sum(self._y))
            classifier = XGBClassifier(
                max_depth=max_depth,
                learning_rate=learning_rate,
                n_estimators=n_estimators,
                n_jobs=self._algorithm_params["n_threads"],
                colsample_bytree=colsample_bytree,
                reg_alpha=self._algorithm_params["reg_alpha"],
                reg_lambda=self._algorithm_params["reg_lambda"],
                scale_pos_weight=scale_pos_weight,
            )
        else:
            classifier = XGBClassifier(
                max_depth=max_depth,
                learning_rate=learning_rate,
                n_estimators=n_estimators,
                n_jobs=self._algorithm_params["n_threads"],
                colsample_bytree=colsample_bytree,
                reg_alpha=self._algorithm_params["reg_alpha"],
                reg_lambda=self._algorithm_params["reg_lambda"],
            )

        classifier.fit(x_train, y_train)
        y_hat_train = classifier.predict(x_train)
        y_hat = classifier.predict(x_test)
        proba_test = classifier.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return classifier, y_hat, auc, y_hat_train

    def _grid_search(
        self,
        x_train,
        x_test,
        y_train,
        y_test,
        max_depth,
        learning_rate,
        n_estimators,
        colsample_bytree,
    ):

        _, y_hat, _, _ = self._launch_xgboost(
            x_train,
            x_test,
            y_train,
            y_test,
            max_depth,
            learning_rate,
            n_estimators,
            colsample_bytree,
        )
        res = utils.evaluate_prediction(y_test, y_hat)

        return res["balanced_accuracy"]

    def _select_best_parameter(self, async_result):

        params_list = []
        accuracies = []

        all_params_acc = []

        for fold in async_result.keys():
            best_params = None
            best_acc = -1

            for params, async_acc in async_result[fold].items():
                acc = async_acc.get()
                if acc > best_acc:
                    best_params = params
                    best_acc = acc

                all_params_acc.append(
                    pd.DataFrame(
                        {
                            "max_depth": params[0],
                            "learning_rate": params[1],
                            "n_estimators": params[2],
                            "colsample_bytree": params[3],
                            "balanced_accuracy": acc,
                        },
                        index=[
                            "i",
                        ],
                    )
                )

            params_list.append(best_params)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)

        best_max_depth = int(round(np.mean([x[0] for x in params_list])))
        best_learning_rate = np.mean([x[1] for x in params_list])
        best_n_estimators = int(round(np.mean([x[2] for x in params_list])))
        best_colsample_bytree = np.mean([x[3] for x in params_list])

        return {
            "max_depth": best_max_depth,
            "learning_rate": best_learning_rate,
            "n_estimators": best_n_estimators,
            "colsample_bytree": best_colsample_bytree,
            "balanced_accuracy": best_acc,
        }

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params["n_threads"])
        async_result = {}
        for i in range(self._algorithm_params["grid_search_folds"]):
            async_result[i] = {}

        x_train = self._x[train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(
            n_splits=self._algorithm_params["grid_search_folds"], shuffle=True
        )
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        parameters_combinations = list(
            itertools.product(
                self._algorithm_params["max_depth_range"],
                self._algorithm_params["learning_rate_range"],
                self._algorithm_params["n_estimators_range"],
                self._algorithm_params["colsample_bytree_range"],
            )
        )

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            x_train_inner = x_train[inner_train_index]
            x_test_inner = x_train[inner_test_index]
            y_train_inner = y_train[inner_train_index]
            y_test_inner = y_train[inner_test_index]

            for parameters in parameters_combinations:
                async_result[i][parameters] = inner_pool.apply_async(
                    self._grid_search,
                    (
                        x_train_inner,
                        x_test_inner,
                        y_train_inner,
                        y_test_inner,
                        parameters[0],
                        parameters[1],
                        parameters[2],
                        parameters[3],
                    ),
                )
        inner_pool.close()
        inner_pool.join()
        best_parameter = self._select_best_parameter(async_result)
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_xgboost(
            x_train,
            x_test,
            y_train,
            y_test,
            best_parameter["max_depth"],
            best_parameter["learning_rate"],
            best_parameter["n_estimators"],
            best_parameter["colsample_bytree"],
        )

        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction(y_test, y_hat)
        result["evaluation_train"] = utils.evaluate_prediction(y_train, y_hat_train)
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index
        result["auc"] = auc

        return result

    def evaluate_no_cv(self, train_index, test_index):

        x_train = self._x[train_index]
        y_train = self._y[train_index]
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        best_parameter = dict()
        best_parameter["max_depth"] = self._algorithm_params["max_depth_range"]
        best_parameter["learning_rate"] = self._algorithm_params["learning_rate_range"]
        best_parameter["n_estimators"] = self._algorithm_params["n_estimators_range"]
        best_parameter["colsample_bytree"] = self._algorithm_params[
            "colsample_bytree_range"
        ]

        _, y_hat, auc, y_hat_train = self._launch_xgboost(
            x_train,
            x_test,
            y_train,
            y_test,
            self._algorithm_params["max_depth_range"],
            self._algorithm_params["learning_rate_range"],
            self._algorithm_params["n_estimators_range"],
            self._algorithm_params["colsample_bytree_range"],
        )
        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction(y_test, y_hat)
        best_parameter["balanced_accuracy"] = result["evaluation"]["balanced_accuracy"]
        result["evaluation_train"] = utils.evaluate_prediction(y_train, y_hat_train)
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index
        result["auc"] = auc

        return result

    def apply_best_parameters(self, results_list):

        mean_bal_acc = np.mean(
            [result["best_parameter"]["balanced_accuracy"] for result in results_list]
        )

        best_max_depth = int(
            round(
                np.mean(
                    [result["best_parameter"]["max_depth"] for result in results_list]
                )
            )
        )
        best_learning_rate = np.mean(
            [result["best_parameter"]["learning_rate"] for result in results_list]
        )
        best_n_estimators = int(
            round(
                np.mean(
                    [
                        result["best_parameter"]["n_estimators"]
                        for result in results_list
                    ]
                )
            )
        )
        best_colsample_bytree = np.mean(
            [result["best_parameter"]["colsample_bytree"] for result in results_list]
        )

        if self._algorithm_params["balanced"]:
            scale_pos_weight = float(len(self._y - sum(self._y)) / sum(self._y))

            classifier = XGBClassifier(
                max_depth=best_max_depth,
                learning_rate=best_learning_rate,
                n_estimators=best_n_estimators,
                n_jobs=self._algorithm_params["n_threads"],
                colsample_bytree=best_colsample_bytree,
                reg_alpha=self._algorithm_params["reg_alpha"],
                reg_lambda=self._algorithm_params["reg_lambda"],
                scale_pos_weight=scale_pos_weight,
            )
        else:
            classifier = XGBClassifier(
                max_depth=best_max_depth,
                learning_rate=best_learning_rate,
                n_estimators=best_n_estimators,
                n_jobs=self._algorithm_params["n_threads"],
                colsample_bytree=best_colsample_bytree,
                reg_alpha=self._algorithm_params["reg_alpha"],
                reg_lambda=self._algorithm_params["reg_lambda"],
            )

        classifier.fit(self._x, self._y)

        return classifier, {
            "max_depth": best_max_depth,
            "learning_rate": best_learning_rate,
            "n_estimators": best_n_estimators,
            "colsample_bytree": best_colsample_bytree,
            "balanced_accuracy": mean_bal_acc,
        }

    def save_classifier(self, classifier, output_dir):
        np.savetxt(
            path.join(output_dir, "feature_importances.txt"),
            classifier.feature_importances_,
        )
        # print classifier.estimators_
        # np.savetxt(path.join(output_dir, 'estimators.txt'), str(classifier.estimators_))

    def save_weights(self, classifier, x, output_dir):

        np.savetxt(
            path.join(output_dir, "weights.txt"), classifier.feature_importances_
        )
        return classifier.feature_importances_

    def save_parameters(self, parameters_dict, output_dir):

        with open(path.join(output_dir, "best_parameters.json"), "w") as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return False

    @staticmethod
    def get_default_parameters():
        parameters_dict = {
            "balanced": False,
            "grid_search_folds": 10,
            "max_depth_range": (0, 6),
            "learning_rate_range": (0.1, 0.3),
            "n_estimators_range": (100, 200),
            "colsample_bytree_range": (0.5, 1),
            "reg_alpha": 0,
            "reg_lambda": 1,
            "n_threads": 15,
        }

        return parameters_dict


class OneVsOneSVM(base.MLAlgorithm):
    def _launch_svc(self, kernel_train, x_test, y_train, y_test, c):

        if self._algorithm_params["balanced"]:
            svc = OneVsOneClassifier(
                SVC(
                    C=c,
                    kernel="precomputed",
                    probability=True,
                    tol=1e-6,
                    class_weight="balanced",
                )
            )
        else:
            svc = OneVsOneClassifier(
                SVC(C=c, kernel="precomputed", probability=True, tol=1e-6)
            )

        svc.fit(kernel_train, y_train)
        y_hat_train = svc.predict(kernel_train)
        y_hat = svc.predict(x_test)
        proba_test = svc.predict_proba(x_test)[:, 1]

        return svc, y_hat, y_hat_train

    def _grid_search(self, kernel_train, x_test, y_train, y_test, c):
        # y_hat is the value predicted

        _, y_hat, _ = self._launch_svc(kernel_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction_multiclass(y_test, y_hat)

        return res["balanced_accuracy"]

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {"c": best_c, "balanced_accuracy": best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params["n_threads"])
        async_result = {}
        for i in range(self._algorithm_params["grid_search_folds"]):
            async_result[i] = {}

        outer_kernel = self._kernel[train_index, :][:, train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(
            n_splits=self._algorithm_params["grid_search_folds"], shuffle=True
        )
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            inner_kernel = outer_kernel[inner_train_index, :][:, inner_train_index]
            x_test_inner = outer_kernel[inner_test_index, :][:, inner_train_index]
            y_train_inner, y_test_inner = (
                y_train[inner_train_index],
                y_train[inner_test_index],
            )

            for c in self._algorithm_params["c_range"]:
                async_result[i][c] = inner_pool.apply_async(
                    self._grid_search,
                    (inner_kernel, x_test_inner, y_train_inner, y_test_inner, c),
                )
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._kernel[test_index, :][:, train_index]
        y_train, y_test = self._y[train_index], self._y[test_index]

        _, y_hat, y_hat_train = self._launch_svc(
            outer_kernel, x_test, y_train, y_test, best_parameter["c"]
        )

        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction_multiclass(y_test, y_hat)
        result["evaluation_train"] = utils.evaluate_prediction_multiclass(
            y_train, y_hat_train
        )
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result["best_parameter"]["c"])
            bal_acc_list.append(result["best_parameter"]["balanced_accuracy"])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params["balanced"]:
            svc = OneVsOneClassifier(
                SVC(
                    C=best_c,
                    kernel="precomputed",
                    probability=True,
                    tol=1e-6,
                    class_weight="balanced",
                )
            )
        else:
            svc = OneVsOneClassifier(
                SVC(C=best_c, kernel="precomputed", probability=True, tol=1e-6)
            )

        svc.fit(self._kernel, self._y)

        return svc, {"c": best_c, "balanced_accuracy": mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(
            path.join(output_dir, "support_vectors_indices.txt"), classifier.support_
        )
        np.savetxt(path.join(output_dir, "intersect.txt"), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        dual_coefficients = classifier.dual_coef_
        sv_indices = classifier.support_

        weighted_sv = dual_coefficients.transpose() * x[sv_indices]
        weights = np.sum(weighted_sv, 0)

        np.savetxt(path.join(output_dir, "weights.txt"), weights)

        return weights

    def save_parameters(self, parameters_dict, output_dir):
        with open(path.join(output_dir, "best_parameters.json"), "w") as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return True

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "balanced": True,
            "grid_search_folds": 10,
            "c_range": np.logspace(-6, 2, 17),
            "n_threads": 15,
        }

        return parameters_dict


class OneVsRestSVM(base.MLAlgorithm):
    def _launch_svc(self, kernel_train, x_test, y_train, y_test, c):

        if self._algorithm_params["balanced"]:
            svc = OneVsRestClassifier(
                SVC(
                    C=c,
                    kernel="precomputed",
                    probability=True,
                    tol=1e-6,
                    class_weight="balanced",
                )
            )
        else:
            svc = OneVsRestClassifier(
                SVC(C=c, kernel="precomputed", probability=True, tol=1e-6)
            )

        svc.fit(kernel_train, y_train)
        y_hat_train = svc.predict(kernel_train)
        y_hat = svc.predict(x_test)
        proba_test = svc.predict_proba(x_test)[:, 1]

        return svc, y_hat, y_hat_train

    def _grid_search(self, kernel_train, x_test, y_train, y_test, c):
        # y_hat is the value predicted

        _, y_hat, _ = self._launch_svc(kernel_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction_multiclass(y_test, y_hat)

        return res["balanced_accuracy"]

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {"c": best_c, "balanced_accuracy": best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params["n_threads"])
        async_result = {}
        for i in range(self._algorithm_params["grid_search_folds"]):
            async_result[i] = {}

        outer_kernel = self._kernel[train_index, :][:, train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(
            n_splits=self._algorithm_params["grid_search_folds"], shuffle=True
        )
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            inner_kernel = outer_kernel[inner_train_index, :][:, inner_train_index]
            x_test_inner = outer_kernel[inner_test_index, :][:, inner_train_index]
            y_train_inner, y_test_inner = (
                y_train[inner_train_index],
                y_train[inner_test_index],
            )

            for c in self._algorithm_params["c_range"]:
                async_result[i][c] = inner_pool.apply_async(
                    self._grid_search,
                    (inner_kernel, x_test_inner, y_train_inner, y_test_inner, c),
                )
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._kernel[test_index, :][:, train_index]
        y_train, y_test = self._y[train_index], self._y[test_index]

        _, y_hat, y_hat_train = self._launch_svc(
            outer_kernel, x_test, y_train, y_test, best_parameter["c"]
        )

        result = dict()
        result["best_parameter"] = best_parameter
        result["evaluation"] = utils.evaluate_prediction_multiclass(y_test, y_hat)
        result["evaluation_train"] = utils.evaluate_prediction_multiclass(
            y_train, y_hat_train
        )
        result["y_hat"] = y_hat
        result["y_hat_train"] = y_hat_train
        result["y"] = y_test
        result["y_train"] = y_train
        result["y_index"] = test_index
        result["x_index"] = train_index

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result["best_parameter"]["c"])
            bal_acc_list.append(result["best_parameter"]["balanced_accuracy"])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params["balanced"]:
            svc = OneVsOneClassifier(
                SVC(
                    C=best_c,
                    kernel="precomputed",
                    probability=True,
                    tol=1e-6,
                    class_weight="balanced",
                )
            )
        else:
            svc = OneVsOneClassifier(
                SVC(C=best_c, kernel="precomputed", probability=True, tol=1e-6)
            )

        svc.fit(self._kernel, self._y)

        return svc, {"c": best_c, "balanced_accuracy": mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(
            path.join(output_dir, "support_vectors_indices.txt"), classifier.support_
        )
        np.savetxt(path.join(output_dir, "intersect.txt"), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        dual_coefficients = classifier.dual_coef_
        sv_indices = classifier.support_

        weighted_sv = dual_coefficients.transpose() * x[sv_indices]
        weights = np.sum(weighted_sv, 0)

        np.savetxt(path.join(output_dir, "weights.txt"), weights)

        return weights

    def save_parameters(self, parameters_dict, output_dir):
        with open(path.join(output_dir, "best_parameters.json"), "w") as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return True

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "balanced": True,
            "grid_search_folds": 10,
            "c_range": np.logspace(-6, 2, 17),
            "n_threads": 15,
        }

        return parameters_dict



# coding: utf8


"""This module contains utilities to handle groups in Clinica.

See CAPS specifications for details about groups.
"""


def check_group_label(group_label):
    """Check that `group_label` is compliant with specifications."""
    if not group_label.isalnum():
        raise ValueError(
            f"Not valid group_label value: it must be composed only by letters "
            f"and/or numbers (given value: {group_label})."
        )


def extract_group_ids(caps_directory):
    """Extract list of group IDs (e.g. ['group-AD', 'group-HC']) based on `caps_directory`/groups folder."""
    import os

    try:
        group_ids = os.listdir(os.path.join(caps_directory, "groups"))
    except FileNotFoundError:
        group_ids = [""]

    return group_ids



# coding: utf8


def zip_nii(in_file: str, same_dir: bool = False):
    import gzip
    import shutil
    from os import getcwd
    from os.path import abspath, join

    from nipype.utils.filemanip import split_filename
    from traits.trait_base import _Undefined

    if (in_file is None) or isinstance(in_file, _Undefined):
        return None

    if not isinstance(in_file, str):  # type(in_file) is list:
        return [zip_nii(f, same_dir) for f in in_file]

    orig_dir, base, ext = split_filename(str(in_file))

    # Already compressed
    if ext[-3:].lower() == ".gz":
        return in_file
    # Not compressed

    out_file = abspath(join(orig_dir if same_dir else getcwd(), base + ext + ".gz"))

    with open(in_file, "rb") as f_in, gzip.open(out_file, "wb") as f_out:
        shutil.copyfileobj(f_in, f_out)

    return out_file


def unzip_nii(in_file: str):
    from nipype.algorithms.misc import Gunzip
    from nipype.utils.filemanip import split_filename
    from traits.trait_base import _Undefined

    if (in_file is None) or isinstance(in_file, _Undefined):
        return None

    if not isinstance(in_file, str):  # type(in_file) is list:
        return [unzip_nii(f) for f in in_file]

    _, base, ext = split_filename(in_file)

    # Not compressed
    if ext[-3:].lower() != ".gz":
        return in_file
    # Compressed
    gunzip = Gunzip(in_file=in_file)
    gunzip.run()
    return gunzip.aggregate_outputs().out_file


def save_participants_sessions(participant_ids, session_ids, out_folder, out_file=None):
    """Save <participant_ids> <session_ids> in <out_folder>/<out_file> TSV file."""
    import os

    import pandas

    from clinica.utils.stream import cprint

    assert len(participant_ids) == len(session_ids)

    os.makedirs(out_folder, exist_ok=True)

    if out_file:
        tsv_file = os.path.join(out_folder, out_file)
    else:
        tsv_file = os.path.join(out_folder, "participants.tsv")

    try:
        data = pandas.DataFrame(
            {
                "participant_id": participant_ids,
                "session_id": session_ids,
            }
        )
        data.to_csv(tsv_file, sep="\t", index=False, encoding="utf-8")
    except Exception as e:
        cprint(msg=f"Impossible to save {out_file} with pandas", lvl="error")
        raise e


def get_subject_id(bids_or_caps_file: str) -> str:
    """Extract "sub-<participant_id>_ses-<session_label>" from BIDS or CAPS file."""
    import re

    m = re.search(r"(sub-[a-zA-Z0-9]+)/(ses-[a-zA-Z0-9]+)", bids_or_caps_file)

    if not m:
        raise ValueError(
            f"Input filename {bids_or_caps_file} is not in a BIDS or CAPS compliant format."
            " It does not contain the subject and session information."
        )

    subject_id = m.group(1) + "_" + m.group(2)

    return subject_id


def get_filename_no_ext(filename):
    """Get filename without extension [".nii.gz", ".tar.gz", ".niml.dset"]."""
    from nipype.utils.filemanip import split_filename

    _, filename_no_ext, _ = split_filename(filename)

    return filename_no_ext


def extract_image_ids(bids_or_caps_files):
    """Extract image IDs (e.g. ['sub-CLNC01_ses-M00', 'sub-CLNC01_ses-M18']  from `bids_or_caps_files`."""
    import re

    id_bids_or_caps_files = [
        re.search(r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)", file).group()
        for file in bids_or_caps_files
    ]
    return id_bids_or_caps_files


def extract_subjects_sessions_from_filename(bids_or_caps_files):
    """Extract subjects/sessions (e.g. ['sub-CLNC01', 'sub-CLNC01']/['ses-M00', 'ses-M18'] from `bids_or_caps_files`."""
    id_bids_or_caps_files = extract_image_ids(bids_or_caps_files)
    split = [image_id.split("_") for image_id in id_bids_or_caps_files]
    subject_ids = [p_id[0] for p_id in split]
    session_ids = [s_id[1] for s_id in split]
    return subject_ids, session_ids


def extract_crash_files_from_log_file(filename):
    """Extract crash files (*.pklz) from `filename`."""
    import os
    import re

    assert os.path.isfile(
        filename
    ), f"extract_crash_files_from_log_file: filename parameter is not a file ({filename})"

    log_file = open(filename, "r")
    crash_files = []
    for line in log_file:
        if re.match("(.*)crashfile:(.*)", line):
            crash_files.append(line.replace("\t crashfile:", "").replace("\n", ""))

    return crash_files


def read_participant_tsv(tsv_file):
    """Extract participant IDs and session IDs from TSV file.

    Raise:
        ClinicaException if tsv_file is not a file
        ClinicaException if participant_id or session_id column is missing from TSV file
    """
    import os

    import pandas as pd

    from clinica.utils.exceptions import ClinicaException

    if not os.path.isfile(tsv_file):
        raise ClinicaException(
            "The TSV file you gave is not a file.\n"
            "Error explanations:\n"
            f"\t- Clinica expected the following path to be a file: {tsv_file}\n"
            "\t- If you gave relative path, did you run Clinica on the good folder?"
        )
    ss_df = pd.io.parsers.read_csv(tsv_file, sep="\t")
    if "participant_id" not in list(ss_df.columns.values):
        raise ClinicaException(
            f"The TSV file does not contain participant_id column (path: {tsv_file})"
        )
    if "session_id" not in list(ss_df.columns.values):
        raise ClinicaException(
            f"The TSV file does not contain session_id column (path: {tsv_file})"
        )
    participants = list(ss_df.participant_id)
    sessions = list(ss_df.session_id)

    # Remove potential whitespace in participant_id or session_id
    return [sub.strip(" ") for sub in participants], [
        ses.strip(" ") for ses in sessions
    ]



__all__ = ["__version__"]

import pkgutil

# Load the Clinica package version
import sys

__version__ = pkgutil.get_data(__package__, "VERSION").decode("ascii").strip()
version = __version__

# import pkg_resources
# version = pkg_resources.require("Clinica")[0].version
# __version__ = version

# python 3.6 minimum version is required
if sys.version_info < (3, 6):
    print(f"Clinica {__version__} requires Python 3.6")
    sys.exit(1)

# Note: The following lines have been commented for speed purposes.
# from clinica.utils.stream import cprint
# import __builtin__
# __builtin__.cprint = cprint



from typing import List, Optional

import click

from clinica.pipelines import cli_param
from clinica.utils.atlas import T1_VOLUME_ATLASES

pipeline_name = "machinelearning-classification"


@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@click.argument(
    "orig_input_data",
    type=click.Choice(["VoxelBased", "RegionBased"]),
)
@click.argument(
    "image_type",
    type=click.Choice(["T1w", "PET"]),
)
@click.argument(
    "algorithm",
    type=click.Choice(["DualSVM", "LogisticRegression", "RandomForest"]),
)
@click.argument(
    "validation",
    type=click.Choice(["RepeatedHoldOut", "RepeatedKFoldCV"]),
)
@click.argument(
    "subjects_visits_tsv",
    type=click.Path(exists=True, resolve_path=True),
)
@click.argument(
    "diagnoses_tsv",
    type=click.Path(exists=True, resolve_path=True),
)
@click.argument(
    "output_directory", type=click.Path(exists=False, writable=True, resolve_path=True)
)
@cli_param.option_group.pipeline_specific_options
@cli_param.option.acq_label
@cli_param.option.suvr_reference_region
@cli_param.option_group.option(
    "-atlas",
    "--atlas",
    type=click.Choice(T1_VOLUME_ATLASES),
    help="One of the atlases generated by t1-volume or pet-volume pipeline.",
)
@cli_param.option_group.common_pipelines_options
@cli_param.option.n_procs
def cli(
    caps_directory: str,
    group_label: str,
    orig_input_data: str,
    image_type: str,
    algorithm: str,
    validation: str,
    subjects_visits_tsv: str,
    diagnoses_tsv: str,
    output_directory: str,
    acq_label: Optional[str] = None,
    suvr_reference_region: Optional[str] = None,
    atlas: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Classification based on machine learning using scikit-learn.


        GROUP_LABEL is a string defining the group label for the current analysis, which helps you keep track of different analyses.

        The third positional argument defines the type of features for classification. It can be 'RegionBased' or 'VoxelBased'.

        The fourth positional argument defines the studied modality ('T1w' or 'PET')

        The fifth positional argument defines the algorithm. It can be 'DualSVM', 'LogisticRegression' or 'RandomForest'.

        The sixth positional argument defines the validation method. It can be 'RepeatedHoldOut' or 'RepeatedKFoldCV'.

        SUBJECTS_VISITS_TSV is a TSV file containing the participant_id and the session_id columns

        DIAGNOSES_TSV is a TSV file where the diagnosis for each participant (identified by a participant ID) is reported (e.g. AD, CN). It allows the algorithm to perform the dual classification (between the two labels reported).

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/MachineLearning_Classification/
    """
    from clinica.utils.exceptions import ClinicaException

    from .ml_workflows import (
        RegionBasedRepHoldOutDualSVM,
        RegionBasedRepHoldOutLogisticRegression,
        RegionBasedRepHoldOutRandomForest,
        RegionBasedRepKFoldDualSVM,
        VoxelBasedRepHoldOutDualSVM,
        VoxelBasedRepKFoldDualSVM,
    )

    if image_type == "PET":
        if not acq_label:
            raise ClinicaException(
                "You selected PET inputs without setting --acq_label flag. "
                "Clinica will now exit."
            )
        if not suvr_reference_region:
            raise ClinicaException(
                "You selected PET inputs without setting --suvr_reference_region flag. "
                "Clinica will now exit."
            )

    if orig_input_data == "RegionBased" and not atlas:
        raise ClinicaException(
            "You selected region-based inputs without setting --atlas flag. "
            "Clinica will now exit."
        )

    if algorithm in ["LogisticRegression", "RandomForest"]:
        if orig_input_data != "RegionBased" or validation != "RepeatedHoldOut":
            raise ClinicaException(
                "LogisticRegression or RandomForest algorithm can only work on region-based featured or RepeatedHoldOut algorithm. "
                "Clinica will now exit."
            )

    if (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "DualSVM")
    ):
        pipeline = RegionBasedRepHoldOutDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedKFoldCV")
        and (algorithm == "DualSVM")
    ):
        pipeline = RegionBasedRepKFoldDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "LogisticRegression")
    ):
        pipeline = RegionBasedRepHoldOutLogisticRegression(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "RegionBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "RandomForest")
    ):
        pipeline = RegionBasedRepHoldOutRandomForest(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            atlas=atlas,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "VoxelBased")
        and (validation == "RepeatedHoldOut")
        and (algorithm == "DualSVM")
    ):
        pipeline = VoxelBasedRepHoldOutDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            n_threads=n_procs,
        )
    elif (
        (orig_input_data == "VoxelBased")
        and (validation == "RepeatedKFoldCV")
        and (algorithm == "DualSVM")
    ):
        pipeline = VoxelBasedRepKFoldDualSVM(
            caps_directory=caps_directory,
            subjects_visits_tsv=subjects_visits_tsv,
            diagnoses_tsv=diagnoses_tsv,
            group_label=group_label,
            image_type=image_type,
            output_dir=output_directory,
            acq_label=acq_label,
            suvr_reference_region=suvr_reference_region,
            n_threads=n_procs,
        )
    else:
        raise ClinicaException(
            "Unknown combination of machine learning classification."
        )

    pipeline.run()


if __name__ == "__main__":
    cli()



# coding: utf-8

"""Module for converting AV45 and Florbetaben PET of ADNI."""


def convert_adni_av45_fbb_pet(
    source_dir, csv_dir, dest_dir, conversion_dir, subjs_list=None, mod_to_update=False
):
    """Convert AV-45 and Florbetaben PET images of ADNI into BIDS format.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        conversion_dir: path to the TSV files including the paths to original images
        subjs_list: subjects list
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from clinica.utils.stream import cprint

    if not subjs_list:
        adni_merge_path = path.join(csv_dir, "ADNIMERGE.csv")
        adni_merge = pd.read_csv(adni_merge_path, sep=",", low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint(
        f"Calculating paths of AV45 and Florbetaben PET images. Output will be stored in {conversion_dir}."
    )
    images = compute_av45_fbb_pet_paths(
        source_dir, csv_dir, dest_dir, subjs_list, conversion_dir
    )
    cprint(
        "Paths of AV45 and Florbetaben PET images found. Exporting images into BIDS ..."
    )
    paths_to_bids(images, dest_dir, "av45_fbb", mod_to_update=mod_to_update)
    cprint(msg="AV45 and Florbetaben PET conversion done.", lvl="debug")


def compute_av45_fbb_pet_paths(
    source_dir, csv_dir, dest_dir, subjs_list, conversion_dir
):
    """Compute the paths to the AV45 and Florbetaben PET images and store them in a TSV file.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS

    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        find_image_path,
        get_images_pet,
    )

    pet_amyloid_col = [
        "Phase",
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Original",
        "Tracer",
    ]
    pet_amyloid_df = pd.DataFrame(columns=pet_amyloid_col)
    pet_amyloid_dfs_list = []

    # Loading needed .csv files
    av45qc = pd.read_csv(path.join(csv_dir, "AV45QC.csv"), sep=",", low_memory=False)
    amyqc = pd.read_csv(path.join(csv_dir, "AMYQC.csv"), sep=",", low_memory=False)
    pet_meta_list = pd.read_csv(
        path.join(csv_dir, "PET_META_LIST.csv"), sep=",", low_memory=False
    )

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list["Subject"] == subj]

        if subject_pet_meta.empty:
            continue

        # QC for AV45 PET images for ADNI 1, GO and 2
        av45_qc_subj = av45qc[(av45qc.PASS == 1) & (av45qc.RID == int(subj[-4:]))]

        # QC for Amyloid PET images for ADNI 3
        amy_qc_subj = amyqc[(amyqc.SCANQLTY == 1) & (amyqc.RID == int(subj[-4:]))]
        amy_qc_subj.insert(0, "EXAMDATE", amy_qc_subj.SCANDATE.to_list())

        # Concatenating visits in both QC files
        amyloid_qc_subj = pd.concat(
            [av45_qc_subj, amy_qc_subj], axis=0, ignore_index=True, sort=False
        )

        sequences_preprocessing_step = [
            "AV45 Co-registered, Averaged",
            "FBB Co-registered, Averaged",
        ]
        subj_dfs_list = get_images_pet(
            subj,
            amyloid_qc_subj,
            subject_pet_meta,
            pet_amyloid_col,
            "Amyloid-PET",
            sequences_preprocessing_step,
        )
        if subj_dfs_list:
            pet_amyloid_dfs_list += subj_dfs_list

    if pet_amyloid_dfs_list:
        pet_amyloid_df = pd.concat(pet_amyloid_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1
        ("128_S_2220", "m48"),
        # Several output images
        ("098_S_4275", "m84"),
    ]

    # Removing known exceptions from images to convert
    if not pet_amyloid_df.empty:
        error_ind = pet_amyloid_df.index[
            pet_amyloid_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        pet_amyloid_df.drop(error_ind, inplace=True)

    images = find_image_path(pet_amyloid_df, source_dir, "Amyloid", "I", "Image_ID")
    images.to_csv(
        path.join(conversion_dir, "amyloid_pet_paths.tsv"), sep="\t", index=False
    )

    return images



# coding: utf8

import os
from multiprocessing.pool import ThreadPool
from os import path

import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit

from clinica.pipelines.machine_learning import base


class KFoldCV(base.MLValidation):
    def validate(self, y):

        if not self._validation_params["splits_indices"]:
            skf = StratifiedKFold(
                n_splits=self._validation_params["n_folds"], shuffle=True
            )
            self._validation_params["splits_indices"] = list(
                skf.split(np.zeros(len(y)), y)
            )

        async_pool = ThreadPool(self._validation_params["n_threads"])
        async_result = {}

        for i in range(self._validation_params["n_folds"]):

            train_index, test_index = self._validation_params["splits_indices"][i]
            async_result[i] = async_pool.apply_async(
                self._ml_algorithm.evaluate, (train_index, test_index)
            )

        async_pool.close()
        async_pool.join()

        for i in range(self._validation_params["n_folds"]):
            self._validation_results.append(async_result[i].get())

        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(
            self._validation_results
        )

        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir: str):
        if not self._validation_results:
            raise Exception(
                "No results to save. Method validate() must be run before save_results()."
            )

        subjects_folds = []
        results_folds = []
        container_dir = path.join(output_dir, "folds")

        os.makedirs(container_dir, exist_ok=True)

        for i in range(len(self._validation_results)):
            subjects_df = pd.DataFrame(
                {
                    "y": self._validation_results[i]["y"],
                    "y_hat": self._validation_results[i]["y_hat"],
                    "y_index": self._validation_results[i]["y_index"],
                }
            )
            subjects_df.to_csv(
                path.join(container_dir, "subjects_fold-" + str(i) + ".tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            subjects_folds.append(subjects_df)

            # fmt: off
            results_df = pd.DataFrame(
                {
                    "balanced_accuracy": self._validation_results[i]["evaluation"]["balanced_accuracy"],
                    "auc": self._validation_results[i]["auc"],
                    "accuracy": self._validation_results[i]["evaluation"]["accuracy"],
                    "sensitivity": self._validation_results[i]["evaluation"]["sensitivity"],
                    "specificity": self._validation_results[i]["evaluation"]["specificity"],
                    "ppv": self._validation_results[i]["evaluation"]["ppv"],
                    "npv": self._validation_results[i]["evaluation"]["npv"],
                    "train_balanced_accuracy": self._validation_results[i]["evaluation_train"]["balanced_accuracy"],
                    "train_accuracy": self._validation_results[i]["evaluation_train"]["accuracy"],
                    "train_sensitivity": self._validation_results[i]["evaluation_train"]["sensitivity"],
                    "train_specificity": self._validation_results[i]["evaluation_train"]["specificity"],
                    "train_ppv": self._validation_results[i]["evaluation_train"]["ppv"],
                    "train_npv": self._validation_results[i]["evaluation_train"]["npv"],
                },
                index=["i"],
            )
            # fmt: on

            results_df.to_csv(
                path.join(container_dir, "results_fold-" + str(i) + ".tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            results_folds.append(results_df)

        all_subjects = pd.concat(subjects_folds)
        all_subjects.to_csv(
            path.join(output_dir, "subjects.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        all_results = pd.concat(results_folds)
        all_results.to_csv(
            path.join(output_dir, "results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        mean_results = pd.DataFrame(
            all_results.apply(np.nanmean).to_dict(),
            columns=all_results.columns,
            index=[
                0,
            ],
        )
        mean_results.to_csv(
            path.join(output_dir, "mean_results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        print("Mean results of the classification:")
        print(
            "Balanced accuracy: %s"
            % (mean_results["balanced_accuracy"].to_string(index=False))
        )
        print("specificity: %s" % (mean_results["specificity"].to_string(index=False)))
        print("sensitivity: %s" % (mean_results["sensitivity"].to_string(index=False)))
        print("auc: %s" % (mean_results["auc"].to_string(index=False)))

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "n_folds": 10,
            "n_threads": 15,
            "splits_indices": None,
            "inner_cv": True,
        }

        return parameters_dict


class RepeatedKFoldCV(base.MLValidation):
    def validate(self, y):

        if not self._validation_params["splits_indices"]:
            self._validation_params["splits_indices"] = []

            for i in range(self._validation_params["n_iterations"]):
                skf = StratifiedKFold(
                    n_splits=self._validation_params["n_folds"], shuffle=True
                )
                self._validation_params["splits_indices"].append(
                    list(skf.split(np.zeros(len(y)), y))
                )

        async_pool = ThreadPool(self._validation_params["n_threads"])
        async_result = {}

        for i in range(self._validation_params["n_iterations"]):

            train_index, test_index = self._validation_params["splits_indices"][i]
            async_result[i] = async_pool.apply_async(
                self._ml_algorithm.evaluate, (train_index, test_index)
            )

        for r in range(self._validation_params["n_iterations"]):

            async_result[r] = {}
            self._validation_results.append([])

            for i in range(self._validation_params["n_folds"]):

                train_index, test_index = self._validation_params["splits_indices"][r][
                    i
                ]
                async_result[r][i] = async_pool.apply_async(
                    self._ml_algorithm.evaluate, (train_index, test_index)
                )

        async_pool.close()
        async_pool.join()
        for r in range(self._validation_params["n_iterations"]):
            for i in range(self._validation_params["n_folds"]):
                self._validation_results[r].append(async_result[r][i].get())

        # TODO Find a better way to estimate best parameter
        flat_results = [result for fold in self._validation_results for result in fold]
        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(
            flat_results
        )

        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir: str):
        if not self._validation_results:
            raise Exception(
                "No results to save. Method validate() must be run before save_results()."
            )

        all_results_list = []
        all_subjects_list = []

        for iteration in range(len(self._validation_results)):

            iteration_dir = path.join(output_dir, "iteration-" + str(iteration))
            os.makedirs(iteration_dir, exist_ok=True)

            iteration_subjects_list = []
            iteration_results_list = []
            folds_dir = path.join(iteration_dir, "folds")

            os.makedirs(folds_dir, exist_ok=True)

            for i in range(len(self._validation_results[iteration])):
                subjects_df = pd.DataFrame(
                    {
                        "y": self._validation_results[iteration][i]["y"],
                        "y_hat": self._validation_results[iteration][i]["y_hat"],
                        "y_index": self._validation_results[iteration][i]["y_index"],
                    }
                )
                subjects_df.to_csv(
                    path.join(folds_dir, "subjects_fold-" + str(i) + ".tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )
                iteration_subjects_list.append(subjects_df)

                # fmt: off
                results_df = pd.DataFrame(
                    {
                        "balanced_accuracy": self._validation_results[iteration][i]["evaluation"]["balanced_accuracy"],
                        "auc": self._validation_results[iteration][i]["auc"],
                        "accuracy": self._validation_results[iteration][i]["evaluation"]["accuracy"],
                        "sensitivity": self._validation_results[iteration][i]["evaluation"]["sensitivity"],
                        "specificity": self._validation_results[iteration][i]["evaluation"]["specificity"],
                        "ppv": self._validation_results[iteration][i]["evaluation"]["ppv"],
                        "npv": self._validation_results[iteration][i]["evaluation"]["npv"],
                        "train_balanced_accuracy": self._validation_results[iteration][i]["evaluation_train"]["balanced_accuracy"],
                        "train_accuracy": self._validation_results[iteration][i]["evaluation_train"]["accuracy"],
                        "train_sensitivity": self._validation_results[iteration][i]["evaluation_train"]["sensitivity"],
                        "train_specificity": self._validation_results[iteration][i]["evaluation_train"]["specificity"],
                        "train_ppv": self._validation_results[iteration][i]["evaluation_train"]["ppv"],
                        "train_npv": self._validation_results[iteration][i]["evaluation_train"]["npv"],
                    },
                    index=["i"],
                )
                # fmt: on

                results_df.to_csv(
                    path.join(folds_dir, "results_fold-" + str(i) + ".tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )
                iteration_results_list.append(results_df)

            iteration_subjects_df = pd.concat(iteration_subjects_list)
            iteration_subjects_df.to_csv(
                path.join(iteration_dir, "subjects.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            all_subjects_list.append(iteration_subjects_df)

            iteration_results_df = pd.concat(iteration_results_list)
            iteration_results_df.to_csv(
                path.join(iteration_dir, "results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )

            mean_results_df = pd.DataFrame(
                iteration_results_df.apply(np.nanmean).to_dict(),
                columns=iteration_results_df.columns,
                index=[
                    0,
                ],
            )
            mean_results_df.to_csv(
                path.join(iteration_dir, "mean_results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            all_results_list.append(mean_results_df)

        all_subjects_df = pd.concat(all_subjects_list)
        all_subjects_df.to_csv(
            path.join(output_dir, "subjects.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        all_results_df = pd.concat(all_results_list)
        all_results_df.to_csv(
            path.join(output_dir, "results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        mean_results_df = pd.DataFrame(
            all_results_df.apply(np.nanmean).to_dict(),
            columns=all_results_df.columns,
            index=[
                0,
            ],
        )
        mean_results_df.to_csv(
            path.join(output_dir, "mean_results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        print("Mean results of the classification:")
        print(
            "Balanced accuracy: %s"
            % (mean_results_df["balanced_accuracy"].to_string(index=False))
        )
        print(
            "specificity: %s" % (mean_results_df["specificity"].to_string(index=False))
        )
        print(
            "sensitivity: %s" % (mean_results_df["sensitivity"].to_string(index=False))
        )
        print("auc: %s" % (mean_results_df["auc"].to_string(index=False)))

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "n_iterations": 100,
            "n_folds": 10,
            "n_threads": 15,
            "splits_indices": None,
            "inner_cv": True,
        }

        return parameters_dict


class RepeatedHoldOut(base.MLValidation):
    def validate(self, y):

        if not self._validation_params["splits_indices"]:
            splits = StratifiedShuffleSplit(
                n_splits=self._validation_params["n_iterations"],
                test_size=self._validation_params["test_size"],
            )
            self._validation_params["splits_indices"] = list(
                splits.split(np.zeros(len(y)), y)
            )

        async_pool = ThreadPool(self._validation_params["n_threads"])
        async_result = {}

        for i in range(self._validation_params["n_iterations"]):

            train_index, test_index = self._validation_params["splits_indices"][i]
            if self._validation_params["inner_cv"]:
                async_result[i] = async_pool.apply_async(
                    self._ml_algorithm.evaluate, (train_index, test_index)
                )
            else:
                async_result[i] = async_pool.apply_async(
                    self._ml_algorithm.evaluate_no_cv, (train_index, test_index)
                )

        async_pool.close()
        async_pool.join()

        for i in range(self._validation_params["n_iterations"]):
            self._validation_results.append(async_result[i].get())

        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(
            self._validation_results
        )
        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir: str):
        if not self._validation_results:
            raise Exception(
                "No results to save. Method validate() must be run before save_results()."
            )

        all_results_list = []
        all_train_subjects_list = []
        all_test_subjects_list = []

        for iteration in range(len(self._validation_results)):

            iteration_dir = path.join(output_dir, "iteration-" + str(iteration))
            os.makedirs(iteration_dir, exist_ok=True)
            iteration_train_subjects_df = pd.DataFrame(
                {
                    "iteration": iteration,
                    "y": self._validation_results[iteration]["y_train"],
                    "y_hat": self._validation_results[iteration]["y_hat_train"],
                    "subject_index": self._validation_results[iteration]["x_index"],
                }
            )
            iteration_train_subjects_df.to_csv(
                path.join(iteration_dir, "train_subjects.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            all_train_subjects_list.append(iteration_train_subjects_df)

            iteration_test_subjects_df = pd.DataFrame(
                {
                    "iteration": iteration,
                    "y": self._validation_results[iteration]["y"],
                    "y_hat": self._validation_results[iteration]["y_hat"],
                    "subject_index": self._validation_results[iteration]["y_index"],
                }
            )
            iteration_test_subjects_df.to_csv(
                path.join(iteration_dir, "test_subjects.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            all_test_subjects_list.append(iteration_test_subjects_df)

            # fmt: off
            iteration_results_df = pd.DataFrame(
                {
                    "balanced_accuracy": self._validation_results[iteration]["evaluation"]["balanced_accuracy"],
                    "auc": self._validation_results[iteration]["auc"],
                    "accuracy": self._validation_results[iteration]["evaluation"]["accuracy"],
                    "sensitivity": self._validation_results[iteration]["evaluation"]["sensitivity"],
                    "specificity": self._validation_results[iteration]["evaluation"]["specificity"],
                    "ppv": self._validation_results[iteration]["evaluation"]["ppv"],
                    "npv": self._validation_results[iteration]["evaluation"]["npv"],
                    "train_balanced_accuracy": self._validation_results[iteration]["evaluation_train"]["balanced_accuracy"],
                    "train_accuracy": self._validation_results[iteration]["evaluation_train"]["accuracy"],
                    "train_sensitivity": self._validation_results[iteration]["evaluation_train"]["sensitivity"],
                    "train_specificity": self._validation_results[iteration]["evaluation_train"]["specificity"],
                    "train_ppv": self._validation_results[iteration]["evaluation_train"]["ppv"],
                    "train_npv": self._validation_results[iteration]["evaluation_train"]["npv"],
                },
                index=["i"],
            )
            # fmt: on

            iteration_results_df.to_csv(
                path.join(iteration_dir, "results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )

            # mean_results_df = pd.DataFrame(iteration_results_df.apply(np.nanmean).to_dict(),
            #                                columns=iteration_results_df.columns, index=[0, ])
            # mean_results_df.to_csv(path.join(iteration_dir, 'mean_results.tsv'),
            #                        index=False, sep='\t', encoding='utf-8')
            all_results_list.append(iteration_results_df)

        all_train_subjects_df = pd.concat(all_train_subjects_list)
        all_train_subjects_df.to_csv(
            path.join(output_dir, "train_subjects.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        all_test_subjects_df = pd.concat(all_test_subjects_list)
        all_test_subjects_df.to_csv(
            path.join(output_dir, "test_subjects.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        all_results_df = pd.concat(all_results_list)
        all_results_df.to_csv(
            path.join(output_dir, "results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        mean_results_df = pd.DataFrame(
            all_results_df.apply(np.nanmean).to_dict(),
            columns=all_results_df.columns,
            index=[
                0,
            ],
        )
        mean_results_df.to_csv(
            path.join(output_dir, "mean_results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        print("Mean results of the classification:")
        print(
            "Balanced accuracy: %s"
            % (mean_results_df["balanced_accuracy"].to_string(index=False))
        )
        print(
            "specificity: %s" % (mean_results_df["specificity"].to_string(index=False))
        )
        print(
            "sensitivity: %s" % (mean_results_df["sensitivity"].to_string(index=False))
        )
        print("auc: %s" % (mean_results_df["auc"].to_string(index=False)))

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "n_iterations": 100,
            "test_size": 0.2,
            "n_threads": 15,
            "splits_indices": None,
            "inner_cv": True,
        }

        return parameters_dict


class LearningCurveRepeatedHoldOut(base.MLValidation):
    def validate(self, y):

        if not self._validation_params["splits_indices"]:
            splits = StratifiedShuffleSplit(
                n_splits=self._validation_params["n_iterations"],
                test_size=self._validation_params["test_size"],
            )
            self._validation_params["splits_indices"] = list(
                splits.split(np.zeros(len(y)), y)
            )

        async_pool = ThreadPool(self._validation_params["n_threads"])
        async_result = {}

        for i in range(self._validation_params["n_iterations"]):
            train_index, test_index = self._validation_params["splits_indices"][i]
            async_result[i] = {}

            skf = StratifiedKFold(
                n_splits=self._validation_params["n_learning_points"], shuffle=False
            )
            inner_cv_splits = list(
                skf.split(np.zeros(len(y[train_index])), y[train_index])
            )

            for j in range(self._validation_params["n_learning_points"]):
                inner_train_index = np.concatenate(
                    [indexes[1] for indexes in inner_cv_splits[: j + 1]]
                ).ravel()
                async_result[i][j] = async_pool.apply_async(
                    self._ml_algorithm.evaluate,
                    (train_index[inner_train_index], test_index),
                )

        async_pool.close()
        async_pool.join()

        for j in range(self._validation_params["n_learning_points"]):
            learning_point_results = []
            for i in range(self._validation_params["n_iterations"]):
                learning_point_results.append(async_result[i][j].get())

            self._validation_results.append(learning_point_results)

        self._classifier = []
        self._best_params = []
        for j in range(self._validation_params["n_learning_points"]):
            classifier, best_params = self._ml_algorithm.apply_best_parameters(
                self._validation_results[j]
            )
            self._classifier.append(classifier)
            self._best_params.append(best_params)

        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir: str):
        if not self._validation_results:
            raise Exception(
                "No results to save. Method validate() must be run before save_results()."
            )

        for learning_point in range(self._validation_params["n_learning_points"]):

            all_results_list = []
            all_subjects_list = []

            learning_point_dir = path.join(
                output_dir, "learning_split-" + str(learning_point)
            )

            for iteration in range(self._validation_params["n_iterations"]):

                iteration_dir = path.join(
                    learning_point_dir, "iteration-" + str(iteration)
                )
                os.makedirs(iteration_dir, exist_ok=True)
                iteration_subjects_df = pd.DataFrame(
                    {
                        "y": self._validation_results[learning_point][iteration]["y"],
                        "y_hat": self._validation_results[learning_point][iteration][
                            "y_hat"
                        ],
                        "y_index": self._validation_results[learning_point][iteration][
                            "y_index"
                        ],
                    }
                )
                iteration_subjects_df.to_csv(
                    path.join(iteration_dir, "subjects.tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )
                all_subjects_list.append(iteration_subjects_df)

                # fmt: off
                iteration_results_df = pd.DataFrame(
                    {
                        "balanced_accuracy": self._validation_results[learning_point][iteration]["evaluation"]["balanced_accuracy"],
                        "auc": self._validation_results[learning_point][iteration]["auc"],
                        "accuracy": self._validation_results[learning_point][iteration]["evaluation"]["accuracy"],
                        "sensitivity": self._validation_results[learning_point][iteration]["evaluation"]["sensitivity"],
                        "specificity": self._validation_results[learning_point][iteration]["evaluation"]["specificity"],
                        "ppv": self._validation_results[learning_point][iteration]["evaluation"]["ppv"],
                        "npv": self._validation_results[learning_point][iteration]["evaluation"]["npv"],
                        "train_balanced_accuracy": self._validation_results[learning_point][iteration]["evaluation_train"]["balanced_accuracy"],
                        "train_accuracy": self._validation_results[learning_point][iteration]["evaluation_train"]["accuracy"],
                        "train_sensitivity": self._validation_results[learning_point][iteration]["evaluation_train"]["sensitivity"],
                        "train_specificity": self._validation_results[learning_point][iteration]["evaluation_train"]["specificity"],
                        "train_ppv": self._validation_results[learning_point][iteration]["evaluation_train"]["ppv"],
                        "train_npv": self._validation_results[learning_point][iteration]["evaluation_train"]["npv"],
                    },
                    index=["i"],
                )
                # fmt: on

                iteration_results_df.to_csv(
                    path.join(iteration_dir, "results.tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )

                mean_results_df = pd.DataFrame(
                    iteration_results_df.apply(np.nanmean).to_dict(),
                    columns=iteration_results_df.columns,
                    index=[
                        0,
                    ],
                )
                mean_results_df.to_csv(
                    path.join(iteration_dir, "mean_results.tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )
                all_results_list.append(mean_results_df)

            all_subjects_df = pd.concat(all_subjects_list)
            all_subjects_df.to_csv(
                path.join(learning_point_dir, "subjects.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )

            all_results_df = pd.concat(all_results_list)
            all_results_df.to_csv(
                path.join(learning_point_dir, "results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )

            mean_results_df = pd.DataFrame(
                all_results_df.apply(np.nanmean).to_dict(),
                columns=all_results_df.columns,
                index=[
                    0,
                ],
            )
            mean_results_df.to_csv(
                path.join(learning_point_dir, "mean_results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )

    @staticmethod
    def get_default_parameters():

        parameters_dict = {
            "n_iterations": 100,
            "test_size": 0.2,
            "n_learning_points": 10,
            "n_threads": 15,
            "splits_indices": None,
            "inner_cv": True,
        }

        return parameters_dict


class RepeatedKFoldCV_Multiclass(base.MLValidation):
    def __init__(self, ml_algorithm):
        self._ml_algorithm = ml_algorithm
        self._repeated_validation_results = []
        self._classifier = None
        self._best_params = None
        self._cv = None

    def validate(self, y, n_iterations=100, n_folds=10, n_threads=15):

        async_pool = ThreadPool(self._validation_params["n_threads"])
        async_result = {}
        self._cv = []

        for r in range(n_iterations):
            skf = StratifiedKFold(n_splits=n_folds, shuffle=True)
            self._cv.append(list(skf.split(np.zeros(len(y)), y)))
            async_result[r] = {}
            self._repeated_validation_results.append([])

            for i in range(n_folds):

                train_index, test_index = self._cv[r][i]
                async_result[r][i] = async_pool.apply_async(
                    self._ml_algorithm.evaluate, (train_index, test_index)
                )

        async_pool.close()
        async_pool.join()
        for r in range(n_iterations):
            for i in range(n_folds):
                self._repeated_validation_results[r].append(async_result[r][i].get())

        # TODO Find a better way to estimate best parameter
        flat_results = [
            result for fold in self._repeated_validation_results for result in fold
        ]
        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(
            flat_results
        )

        return self._classifier, self._best_params, self._repeated_validation_results

    def save_results(self, output_dir: str):
        if not self._repeated_validation_results:
            raise Exception(
                "No results to save. Method validate() must be run before save_results()."
            )

        all_results_list = []
        all_subjects_list = []

        for iteration in range(len(self._repeated_validation_results)):

            iteration_dir = path.join(output_dir, "iteration-" + str(iteration))
            os.makedirs(iteration_dir, exist_ok=True)

            iteration_subjects_list = []
            iteration_results_list = []
            folds_dir = path.join(iteration_dir, "folds")

            os.makedirs(folds_dir, exist_ok=True)

            for i in range(len(self._repeated_validation_results[iteration])):
                subjects_df = pd.DataFrame(
                    {
                        "y": self._repeated_validation_results[iteration][i]["y"],
                        "y_hat": self._repeated_validation_results[iteration][i][
                            "y_hat"
                        ],
                        "y_index": self._repeated_validation_results[iteration][i][
                            "y_index"
                        ],
                    }
                )
                subjects_df.to_csv(
                    path.join(folds_dir, "subjects_fold-" + str(i) + ".tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )
                iteration_subjects_list.append(subjects_df)

                # fmt: off
                results_df = pd.DataFrame(
                    {
                        "balanced_accuracy": self._repeated_validation_results[iteration][i]["evaluation"]["balanced_accuracy"],
                        "accuracy": self._repeated_validation_results[iteration][i]["evaluation"]["accuracy"],
                        "train_balanced_accuracy": self._repeated_validation_results[iteration][i]["evaluation_train"]["balanced_accuracy"],
                        "train_accuracy": self._repeated_validation_results[iteration][i]["evaluation_train"]["accuracy"],
                    },
                    index=[
                        "i",
                    ],
                )
                # fmt: on

                results_df.to_csv(
                    path.join(folds_dir, "results_fold-" + str(i) + ".tsv"),
                    index=False,
                    sep="\t",
                    encoding="utf-8",
                )
                iteration_results_list.append(results_df)

            iteration_subjects_df = pd.concat(iteration_subjects_list)
            iteration_subjects_df.to_csv(
                path.join(iteration_dir, "subjects.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            all_subjects_list.append(iteration_subjects_df)

            iteration_results_df = pd.concat(iteration_results_list)
            iteration_results_df.to_csv(
                path.join(iteration_dir, "results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )

            mean_results_df = pd.DataFrame(
                iteration_results_df.apply(np.nanmean).to_dict(),
                columns=iteration_results_df.columns,
                index=[
                    0,
                ],
            )
            mean_results_df.to_csv(
                path.join(iteration_dir, "mean_results.tsv"),
                index=False,
                sep="\t",
                encoding="utf-8",
            )
            all_results_list.append(mean_results_df)

        all_subjects_df = pd.concat(all_subjects_list)
        all_subjects_df.to_csv(
            path.join(output_dir, "subjects.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        all_results_df = pd.concat(all_results_list)
        all_results_df.to_csv(
            path.join(output_dir, "results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        mean_results_df = pd.DataFrame(
            all_results_df.apply(np.nanmean).to_dict(),
            columns=all_results_df.columns,
            index=[
                0,
            ],
        )
        mean_results_df.to_csv(
            path.join(output_dir, "mean_results.tsv"),
            index=False,
            sep="\t",
            encoding="utf-8",
        )

        print("Mean results of the classification:")
        print(
            "Balanced accuracy: %s"
            % (mean_results_df["balanced_accuracy"].to_string(index=False))
        )



# coding: utf8

import clinica.pipelines.engine as cpe


class T1VolumeRegisterDartel(cpe.Pipeline):
    """T1VolumeExistingDartel - Reuse existing Dartel template.

    Returns:
        A clinica pipeline object containing the T1VolumeExistingDartel pipeline.
    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        if "group_label" not in self.parameters.keys():
            raise KeyError("Missing compulsory group_label key in pipeline parameter.")
        self.parameters.setdefault("tissues", [1, 2, 3])

        check_group_label(self.parameters["group_label"])

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ["dartel_input_images", "dartel_iteration_templates"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return ["dartel_flow_fields"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.input_files import (
            t1_volume_dartel_input_tissue,
            t1_volume_i_th_iteration_group_template,
        )
        from clinica.utils.inputs import clinica_file_reader, clinica_group_reader
        from clinica.utils.ux import print_images_to_process

        read_input_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )

        all_errors = []

        # Dartel Input Tissues
        # ====================
        d_input = []
        for tissue_number in self.parameters["tissues"]:
            try:
                current_file = clinica_file_reader(
                    self.subjects,
                    self.sessions,
                    self.caps_directory,
                    t1_volume_dartel_input_tissue(tissue_number),
                )
                d_input.append(current_file)
            except ClinicaException as e:
                all_errors.append(e)

        # Dartel Templates
        # ================
        dartel_iter_templates = []
        for i in range(1, 7):
            try:
                current_iter = clinica_group_reader(
                    self.caps_directory,
                    t1_volume_i_th_iteration_group_template(
                        self.parameters["group_label"], i
                    ),
                )

                dartel_iter_templates.append(current_iter)
            except ClinicaException as e:
                all_errors.append(e)

        if len(all_errors) > 0:
            error_message = "Clinica faced error(s) while trying to read files in your CAPS/BIDS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaCAPSError(error_message)

        read_input_node.inputs.dartel_input_images = d_input
        read_input_node.inputs.dartel_iteration_templates = dartel_iter_templates

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)

        # fmt: off
        self.connect(
            [
                (read_input_node, self.input_node, [("dartel_input_images", "dartel_input_images")]),
                (read_input_node, self.input_node, [("dartel_iteration_templates", "dartel_iteration_templates")]),
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import re

        import nipype.interfaces.io as nio
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import zip_nii

        # Writing flowfields into CAPS
        # ============================
        write_flowfields_node = npe.MapNode(
            name="write_flowfields_node",
            iterfield=["container", "flow_fields"],
            interface=nio.DataSink(infields=["flow_fields"]),
        )
        write_flowfields_node.inputs.base_directory = self.caps_directory
        write_flowfields_node.inputs.parameterization = False
        write_flowfields_node.inputs.container = [
            "subjects/"
            + self.subjects[i]
            + "/"
            + self.sessions[i]
            + "/t1/spm/dartel/group-"
            + self.parameters["group_label"]
            for i in range(len(self.subjects))
        ]
        write_flowfields_node.inputs.regexp_substitutions = [
            (r"(.*)_Template(\.nii(\.gz)?)$", r"\1\2"),
            (r"(.*)c1(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-graymatter\3"),
            (r"(.*)c2(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-whitematter\3"),
            (r"(.*)c3(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-csf\3"),
            (r"(.*)c4(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-bone\3"),
            (r"(.*)c5(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-softtissue\3"),
            (r"(.*)c6(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-background\3"),
            (r"(.*)r(sub-.*)(\.nii(\.gz)?)$", r"\1\2\3"),
            (r"(.*)_dartelinput(\.nii(\.gz)?)$", r"\1\2"),
            (
                r"(.*)flow_fields/u_(sub-.*)_segm-.*(\.nii(\.gz)?)$",
                r"\1\2_target-"
                + re.escape(self.parameters["group_label"])
                + r"_transformation-forward_deformation\3",
            ),
            (r"trait_added", r""),
        ]

        # fmt: off
        self.connect(
            [
                (self.output_node, write_flowfields_node, [(("dartel_flow_fields", zip_nii, True), "flow_fields")])
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.t1_volume_register_dartel.t1_volume_register_dartel_utils as utils
        from clinica.utils.filemanip import unzip_nii

        # Unzipping
        # =========
        unzip_dartel_input_node = npe.MapNode(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_dartel_input_node",
            iterfield=["in_file"],
        )
        unzip_templates_node = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_templates_node",
        )
        # DARTEL with existing template
        # =============================
        dartel_existing_template = npe.MapNode(
            utils.DARTELExistingTemplate(),
            name="dartel_existing_template",
            iterfield=["image_files"],
        )

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, unzip_dartel_input_node, [("dartel_input_images", "in_file")]),
                (self.input_node, unzip_templates_node, [("dartel_iteration_templates", "in_file")]),
                (unzip_dartel_input_node, dartel_existing_template, [(("out_file", utils.prepare_dartel_input_images), "image_files")]),
                (unzip_templates_node, dartel_existing_template, [(("out_file", utils.create_iteration_parameters, None), "iteration_parameters")]),
                (dartel_existing_template, self.output_node, [("dartel_flow_fields", "dartel_flow_fields")]),
            ]
        )
        # fmt: on



import click

from clinica.iotools.converters import cli_param


@click.command(name="aibl-to-bids")
@cli_param.dataset_directory
@cli_param.clinical_data_directory
@cli_param.bids_directory
@cli_param.clinical_data_only
@click.option(
    "--overwrite",
    is_flag=True,
    help="Overwrites previously written nifti and json files.",
)
def cli(
    dataset_directory: str,
    clinical_data_directory: str,
    bids_directory: str,
    clinical_data_only: bool = False,
    overwrite: bool = False,
) -> None:
    """AIBL to BIDS converter.

    Convert the imaging and clinical data of AIBL (https://aibl.csiro.au/adni/index.html), located in DATASET_DIRECTORY
    and CLINICAL_DATA_DIRECTORY respectively, to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from os import makedirs

    from clinica.iotools.converters.aibl_to_bids.aibl_to_bids import (
        convert_clinical_data,
        convert_images,
    )
    from clinica.utils.check_dependency import check_dcm2niix

    check_dcm2niix()

    makedirs(bids_directory, exist_ok=True)

    if not clinical_data_only:
        convert_images(
            dataset_directory,
            clinical_data_directory,
            bids_directory,
            overwrite,
        )

    convert_clinical_data(bids_directory, clinical_data_directory)



# coding: utf8
from typing import List, Optional


def extract_slices(input_tensor, slice_direction=0, slice_mode="single"):
    """Extracts the slices from three directions

    This function extracts slices form the preprocesed nifti image.  The
    direction of extraction can be defined either on sagital direction (0),
    cornal direction (1) or axial direction (other). The output slices can be
    stores following two modes: single (1 channel) ou RGB (3 channels, all the
    same).


    Args:
        input_tensor: tensor version of the nifti MRI.
        slice_direction: which axis direction that the slices were extracted
        slice_mode: 'single' or 'RGB'.

    Returns:
        file: multiple tensors saved on the disk, suffixes corresponds to
            indexes of the slices. Same location than input file.
    """
    import os

    import torch

    image_tensor = torch.load(input_tensor)
    # reshape the tensor, delete the first dimension for slice-level
    image_tensor = image_tensor.view(
        image_tensor.shape[1], image_tensor.shape[2], image_tensor.shape[3]
    )

    # sagital
    # M and N correspond to the first and last slices (if need to remove)
    M = 0
    N = 0
    slice_list_sag = range(
        M, image_tensor.shape[0] - N
    )  # delete the first M slices and last N slices

    basedir = os.getcwd()
    input_tensor_filename = os.path.basename(input_tensor)

    txt_idx = input_tensor_filename.rfind("_")
    it_filename_prefix = input_tensor_filename[0:txt_idx]
    it_filename_suffix = input_tensor_filename[txt_idx:]

    output_file_original = []
    output_file_rgb = []
    if slice_direction == 0:
        for index_slice, index_slice_list in zip(
            slice_list_sag, range(len(slice_list_sag))
        ):
            # for i in slice_list:
            # sagital
            slice_select_sag = image_tensor[index_slice, :, :]

            extracted_slice_original_sag = slice_select_sag.unsqueeze(
                0
            )  # shape should be 1 * W * L

            # train for transfer learning, creating the fake RGB image.
            slice_select_sag = (slice_select_sag - slice_select_sag.min()) / (
                slice_select_sag.max() - slice_select_sag.min()
            )
            extracted_slice_rgb_sag = torch.stack(
                (slice_select_sag, slice_select_sag, slice_select_sag)
            )  # shape should be 3 * W * L

            # save into .pt format
            if slice_mode == "single":
                output_file_original.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + "_axis-sag_channel-single_slice-"
                        + str(index_slice)
                        + it_filename_suffix,
                    )
                )
                torch.save(
                    extracted_slice_original_sag.clone(),
                    output_file_original[index_slice_list],
                )
            elif slice_mode == "rgb":
                output_file_rgb.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + "_axis-sag_channel-rgb_slice-"
                        + str(index_slice)
                        + it_filename_suffix,
                    )
                )
                torch.save(
                    extracted_slice_rgb_sag.clone(), output_file_rgb[index_slice_list]
                )

    elif slice_direction == 1:
        # cornal
        slice_list_cor = range(
            M, image_tensor.shape[1] - N
        )  # delete the first M slices and last N slices
        for index_slice, index_slice_list in zip(
            slice_list_cor, range(len(slice_list_cor))
        ):
            # for i in slice_list:
            # sagital
            slice_select_cor = image_tensor[:, index_slice, :]

            extracted_slice_original_cor = slice_select_cor.unsqueeze(
                0
            )  # shape should be 1 * W * L

            # train for transfer learning, creating the fake RGB image.
            slice_select_cor = (slice_select_cor - slice_select_cor.min()) / (
                slice_select_cor.max() - slice_select_cor.min()
            )
            extracted_slice_rgb_cor = torch.stack(
                (slice_select_cor, slice_select_cor, slice_select_cor)
            )  # shape should be 3 * W * L

            # save into .pt format
            if slice_mode == "single":
                output_file_original.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + "_axis-cor_channel-single_slice-"
                        + str(index_slice)
                        + it_filename_suffix,
                    )
                )
                torch.save(
                    extracted_slice_original_cor.clone(),
                    output_file_original[index_slice_list],
                )
            elif slice_mode == "rgb":
                output_file_rgb.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + "_axis-cor_channel-rgb_slice-"
                        + str(index_slice)
                        + it_filename_suffix,
                    )
                )
                torch.save(
                    extracted_slice_rgb_cor.clone(), output_file_rgb[index_slice_list]
                )

    else:

        # axial
        slice_list_axi = range(
            M, image_tensor.shape[2] - N
        )  # delete the first M slices and last N slices
        for index_slice, index_slice_list in zip(
            slice_list_axi, range(len(slice_list_axi))
        ):
            # for i in slice_list:
            # sagital
            slice_select_axi = image_tensor[:, :, index_slice]

            extracted_slice_original_axi = slice_select_axi.unsqueeze(
                0
            )  # shape should be 1 * W * L

            # train for transfer learning, creating the fake RGB image.
            slice_select_axi = (slice_select_axi - slice_select_axi.min()) / (
                slice_select_axi.max() - slice_select_axi.min()
            )
            extracted_slice_rgb_axi = torch.stack(
                (slice_select_axi, slice_select_axi, slice_select_axi)
            )  # shape should be 3 * W * L

            # save into .pt format
            if slice_mode == "single":
                output_file_original.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + "_axis-axi_channel-single_slice-"
                        + str(index_slice)
                        + it_filename_suffix,
                    )
                )
                torch.save(
                    extracted_slice_original_axi.clone(),
                    output_file_original[index_slice_list],
                )
            elif slice_mode == "rgb":
                output_file_rgb.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + "_axis-axi_channel-rgb_slice-"
                        + str(index_slice)
                        + it_filename_suffix,
                    )
                )
                torch.save(
                    extracted_slice_rgb_axi.clone(), output_file_rgb[index_slice_list]
                )

    return output_file_rgb, output_file_original


def extract_patches(input_tensor, patch_size, stride_size):
    """Extracts the patches

    This function extracts patches form the preprocesed nifti image. Patch size
    if provieded as input and also the stride size. If stride size is smaller
    than the patch size an overlap exist between consecutive patches. If stride
    size is equal to path size there is no overlap. Otherwise, unprocessed
    zones can exits.

    Args:
        input_tensor: tensor version of the nifti MRI.
        patch_size: size of a single patch.
        stride_size: size of the stride leading to next patch.

    Returns:
        file: multiple tensors saved on the disk, suffixes corresponds to
            indexes of the patches. Same location than input file.
    """
    import os

    import torch

    basedir = os.getcwd()
    image_tensor = torch.load(input_tensor)

    # use classifiers tensor.upfold to crop the patch.
    patches_tensor = (
        image_tensor.unfold(1, patch_size, stride_size)
        .unfold(2, patch_size, stride_size)
        .unfold(3, patch_size, stride_size)
        .contiguous()
    )
    # the dimension of patch_tensor should be [1, patch_num1, patch_num2, patch_num3, patch_size1, patch_size2, patch_size3]
    patches_tensor = patches_tensor.view(-1, patch_size, patch_size, patch_size)

    input_tensor_filename = os.path.basename(input_tensor)
    txt_idx = input_tensor_filename.rfind("_")
    it_filename_prefix = input_tensor_filename[0:txt_idx]
    it_filename_suffix = input_tensor_filename[txt_idx:]

    output_patch = []
    for index_patch in range(patches_tensor.shape[0]):
        extracted_patch = patches_tensor[index_patch, ...].unsqueeze_(
            0
        )  # add one dimension
        # save into .pt format
        output_patch.append(
            os.path.join(
                basedir,
                it_filename_prefix
                + "_patchsize-"
                + str(patch_size)
                + "_stride-"
                + str(stride_size)
                + "_patch-"
                + str(index_patch)
                + it_filename_suffix,
            )
        )
        torch.save(extracted_patch.clone(), output_patch[index_patch])

    return output_patch


def save_as_pt(input_img):
    """Saves PyTorch tensor version of the nifti image

    This function convert nifti image to tensor (.pt) version of the image.
    Tensor version is saved at the same location than input_img.

    Args:
        input_tensor: tensor version of the nifti MRI.

    Returns:
        filename (str): single tensor file  saved on the disk. Same location than input file.

    """

    import os

    import nibabel as nib
    import torch

    basedir = os.getcwd()
    image_array = nib.load(input_img).get_fdata(dtype="float32")
    image_tensor = torch.from_numpy(image_array).unsqueeze(0).float()
    # make sure the tensor dtype is torch.float32
    output_file = os.path.join(
        basedir, os.path.basename(input_img).split(".nii.gz")[0] + ".pt"
    )
    # save
    torch.save(image_tensor.clone(), output_file)

    return output_file


# ROI extraction utils


def check_mask_list(
    masks_location: str,
    roi_list: List[str],
    mask_pattern: Optional[str] = None,
    cropping: Optional[bool] = None,
):
    import nibabel as nib
    import numpy as np

    for roi in roi_list:
        roi_path, desc = find_mask_path(masks_location, roi, mask_pattern, cropping)
        if not roi_path:
            raise ValueError(
                f"The ROI '{roi}' does not correspond to a mask in the CAPS directory. {desc}"
            )
        roi_mask = nib.load(roi_path).get_fdata()
        mask_values = set(np.unique(roi_mask))
        if mask_values != {0, 1}:
            raise ValueError(
                "The ROI masks used should be binary (composed of 0 and 1 only)."
            )


def find_mask_path(
    masks_location: str,
    roi: str,
    mask_pattern: Optional[str] = None,
    cropping: Optional[bool] = None,
):
    """Finds masks corresponding to the pattern asked and containing the adequate cropping description"""
    from glob import glob
    from os import path

    mask_pattern = mask_pattern or ""

    # Check that pattern begins and ends with _ to avoid mixing keys
    if mask_pattern:
        if not mask_pattern.endswith("_"):
            mask_pattern += "_"
        if not mask_pattern.startswith("_"):
            mask_pattern = "_" + mask_pattern

    candidates_pattern = path.join(
        masks_location, f"*{mask_pattern}*_roi-{roi}_mask.nii*"
    )
    desc = f"The mask should follow the pattern {candidates_pattern} "
    candidates = glob(candidates_pattern)
    if cropping:
        candidates = [mask for mask in candidates if "_desc-Crop_" in mask]
        desc += f"and contain '_desc-Crop_' string."
    else:
        candidates = [mask for mask in candidates if "_desc-Crop_" not in mask]
        desc += f"and not contain '_desc-Crop_' string."

    if len(candidates) == 0:
        return None, desc
    else:
        return min(candidates, key=len), desc


def compute_output_pattern(mask_path, crop_output):
    """
    Computes the output pattern of the region cropped (without the source file prefix)
    Args:
        mask_path: path to the masks
        crop_output: If True the output is cropped, and the descriptor CropRoi must exist
    Returns:
        the output pattern
    """
    from os import path

    mask_filename = path.basename(mask_path)
    template_id = mask_filename.split("_")[0].split("-")[1]
    mask_descriptors = mask_filename.split("_")[1:-2:]
    roi_id = mask_filename.split("_")[-2].split("-")[1]
    if "desc-Crop" not in mask_descriptors and crop_output:
        mask_descriptors = ["desc-CropRoi"] + mask_descriptors
    elif "desc-Crop" in mask_descriptors:
        mask_descriptors = [
            descriptor for descriptor in mask_descriptors if descriptor != "desc-Crop"
        ]
        if crop_output:
            mask_descriptors = ["desc-CropRoi"] + mask_descriptors
        else:
            mask_descriptors = ["desc-CropImage"] + mask_descriptors

    mask_pattern = "_".join(mask_descriptors)
    output_pattern = f"space-{template_id}_{mask_pattern}_roi-{roi_id}"

    return output_pattern


def extract_roi(
    input_tensor,
    masks_location,
    mask_pattern,
    cropped_input,
    roi_list,
    uncrop_output,
):
    """Extracts regions of interest defined by masks
    This function extracts regions of interest from preprocessed nifti images.
    The regions are defined using binary masks that must be located in the CAPS
    at `masks/tpl-<template>`.
    Args:
        input_path: path to the tensor version of the nifti MRI.
        basedir: path to the extracted files.
        masks_location: path to the masks
        mask_pattern: pattern to identify the masks
        cropped_input: if the input is cropped or not (contains desc-Crop)
        roi_list: list of the names of the regions that will be extracted.
        uncrop_output: if True, the final region is not cropped.
    Returns:
        file: multiple tensors saved on the disk, suffixes corresponds to
            indexes of the patches. Same location than input file.
    """
    import os

    import nibabel as nib
    import numpy as np
    import torch

    from clinica.pipelines.deeplearning_prepare_data.deeplearning_prepare_data_utils import (
        compute_output_pattern,
        find_mask_path,
    )

    image_tensor = torch.load(input_tensor)
    basedir = os.getcwd()

    input_tensor_filename = os.path.basename(input_tensor)

    sub_ses_prefix = "_".join(input_tensor_filename.split("_")[0:3:])
    if not sub_ses_prefix.endswith("_T1w"):
        sub_ses_prefix = "_".join(input_tensor_filename.split("_")[0:2:])
    input_suffix = input_tensor_filename.split("_")[-1].split(".")[0]

    output_roi = []
    for index_roi, roi in enumerate(roi_list):
        # read mask
        mask_path, _ = find_mask_path(masks_location, roi, mask_pattern, cropped_input)
        mask_np = nib.load(mask_path).get_fdata()
        if len(mask_np.shape) == 3:
            mask_np = mask_np[np.newaxis, :]

        extracted_roi = image_tensor * mask_np
        if not uncrop_output:
            extracted_roi = extracted_roi[
                np.ix_(
                    mask_np.any((1, 2, 3)),
                    mask_np.any((0, 2, 3)),
                    mask_np.any((0, 1, 3)),
                    mask_np.any((0, 1, 2)),
                )
            ]
        extracted_roi = extracted_roi.float()
        # save into .pt format
        output_pattern = compute_output_pattern(mask_path, not uncrop_output)
        output_roi.append(
            os.path.join(
                basedir, f"{sub_ses_prefix}_{output_pattern}_{input_suffix}.pt"
            )
        )
        # os.makedirs(basedir, exist_ok=True)
        torch.save(extracted_roi.clone(), output_roi[index_roi])

    return output_roi


TEMPLATE_DICT = {
    "t1-linear": "MNI152NLin2009cSym",
    "t1-extensive": "Ixi549Space",
    "pet-linear": "MNI152NLin2009cSym",
}



# coding: utf8

"""
Convert the AIBL dataset (http://www.aibl.csiro.au/) into BIDS.
"""


def convert_images(path_to_dataset, path_to_csv, bids_dir, overwrite=False):

    # Conversion of the entire dataset in BIDS
    from os.path import exists

    from clinica.iotools.converters.aibl_to_bids.aibl_utils import paths_to_bids
    from clinica.utils.stream import cprint

    list_of_created_files = []

    for modality in ["t1", "av45", "flute", "pib"]:
        list_of_created_files.append(
            paths_to_bids(
                path_to_dataset, path_to_csv, bids_dir, modality, overwrite=overwrite
            )
        )

    error_string = ""
    for modality_list in list_of_created_files:
        for file in modality_list:
            if not exists(str(file)):
                error_string = error_string + str(file) + "\n"
    if error_string != "":
        cprint(msg=f"The following file were not converted: {error_string}", lvl="warning")


def convert_clinical_data(bids_dir, path_to_csv):
    # clinical specifications in BIDS
    from os.path import exists, join, realpath, split

    import clinica.iotools.bids_utils as bids
    from clinica.iotools.converters.aibl_to_bids.aibl_utils import (
        create_participants_df_AIBL,
        create_scans_dict_AIBL,
        create_sessions_dict_AIBL,
    )
    from clinica.utils.stream import cprint

    clinical_spec_path = join(
        split(realpath(__file__))[0], "../../data/clinical_specifications"
    )
    # if not exists(clinical_spec_path):
    #    raise FileNotFoundError(
    #        f"{clinical_spec_path} file not found ! This is an internal file of Clinica."
    #    )

    cprint("Creating modality agnostic files...")
    bids.write_modality_agnostic_files("AIBL", bids_dir)

    cprint("Creating participants.tsv...")
    create_participants_df_AIBL(
        bids_dir, clinical_spec_path, path_to_csv, delete_non_bids_info=True
    )

    cprint("Creating sessions files...")
    create_sessions_dict_AIBL(bids_dir, path_to_csv, clinical_spec_path)

    cprint("Creating scans files...")
    create_scans_dict_AIBL(bids_dir, path_to_csv, clinical_spec_path)
